# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import sqlite3
import itertools
from urllib.parse import urlparse
from threading import Lock

from eolie.utils import noaccents, get_random_string
from eolie.define import EOLIE_DATA_PATH, Type
from eolie.localized import LocalizedCollation
from eolie.sqlcursor import SqlCursor
from eolie.logger import Logger
from eolie.database_upgrade import DatabaseUpgrade


class DatabaseHistory:
    """
        Eolie history db
    """
    DB_PATH = "%s/history.db" % EOLIE_DATA_PATH

    __UPGRADES = {
        1: "ALTER TABLE history ADD opened INT NOT NULL DEFAULT 0",
        2: "ALTER TABLE history ADD netloc TEXT NOT NULL DEFAULT ''",
        3: "DELETE FROM history WHERE popularity=0",
        4: "DELETE FROM history_atime WHERE NOT EXISTS (SELECT * FROM history\
            WHERE history.rowid=history_atime.history_id)"
    }

    # SQLite documentation:
    # In SQLite, a column with type INTEGER PRIMARY KEY
    # is an alias for the ROWID.
    # Here, we define an id INT PRIMARY KEY but never feed it,
    # this make VACUUM not destroy rowids...
    __create_history = '''CREATE TABLE history (
                                               id INTEGER PRIMARY KEY,
                                               title TEXT NOT NULL,
                                               uri TEXT NOT NULL,
                                               netloc TEXT NOT NULL,
                                               guid TEXT NOT NULL,
                                               mtime REAL NOT NULL,
                                               opened INT NOT NULL DEFAULT 0,
                                               popularity INT NOT NULL
                                               )'''
    __create_history_atime = '''CREATE TABLE history_atime (
                                                history_id INT NOT NULL,
                                                atime REAL NOT NULL
                                               )'''

    __create_history_orderby_idx = """CREATE INDEX
                                               idx_orderby ON history(
                                               mtime, popularity)"""
    __create_history_where_idx = """CREATE INDEX
                                               idx_where ON history(
                                               uri, title)"""

    def __init__(self):
        """
            Create database tables or manage update if needed
        """
        upgrade = DatabaseUpgrade(Type.HISTORY)
        self.thread_lock = Lock()
        if not GLib.file_test(self.DB_PATH, GLib.FileTest.IS_REGULAR):
            try:
                if not GLib.file_test(EOLIE_DATA_PATH, GLib.FileTest.IS_DIR):
                    GLib.mkdir_with_parents(EOLIE_DATA_PATH, 0o0750)
                # Create db schema
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_history)
                    sql.execute(self.__create_history_atime)
                    sql.execute(self.__create_history_orderby_idx)
                    sql.execute(self.__create_history_where_idx)
                    sql.execute("PRAGMA user_version=%s" % upgrade.version)
            except Exception as e:
                Logger.error("DatabaseHistory::__init__(): %s", e)
        else:
            upgrade.upgrade(self)

    def add(self, title, uri, mtime, guid=None, atimes=[]):
        """
            Add a new entry to history, if exists, update it
            @param title as str
            @param uri as str
            @param mtime as int
            @parma guid as str
            @param atime as [int]
            @return history id as int
        """
        if not uri:
            return
        uri = uri.rstrip('/')
        parsed = urlparse(uri)
        with SqlCursor(self, True) as sql:
            result = sql.execute("SELECT rowid, popularity FROM history\
                                  WHERE uri=?", (uri,))
            v = result.fetchone()
            # Update current item
            if v is not None:
                history_id = v[0]
                guid = self.get_guid(history_id)
                sql.execute("UPDATE history\
                             SET uri=?, netloc=?, mtime=?,\
                                 title=?, guid=?, popularity=?\
                             WHERE rowid=?", (uri, parsed.netloc, mtime, title,
                                              guid, v[1] + 1, history_id))
            # Add a new item
            else:
                # Find an uniq guid
                while guid is None:
                    guid = get_random_string(12)
                    if self.exists_guid(guid):
                        guid = None
                result = sql.execute("INSERT INTO history\
                                      (title, uri, netloc,\
                                       mtime, popularity, guid)\
                                      VALUES (?, ?, ?, ?, ?, ?)",
                                     (title, uri, parsed.netloc,
                                      mtime, 0, guid))
                history_id = result.lastrowid
            # Only add new atimes to db
            if not atimes:
                atimes = [mtime]
            current_atimes = self.get_atimes(history_id)
            for atime in atimes:
                if atime not in current_atimes:
                    sql.execute("INSERT INTO history_atime\
                                 (history_id, atime)\
                                 VALUES (?, ?)", (history_id, atime))
            return history_id

    def remove(self, history_id):
        """
            Remove item from history
            @param history id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE from history\
                         WHERE rowid=?", (history_id,))
            sql.execute("DELETE from history_atime\
                         WHERE history_id=?", (history_id,))

    def clear_from(self, atime):
        """
            Clear history from atime
            @param atime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM history_atime\
                         WHERE atime >= ?", (atime,))

    def clear_to(self, atime):
        """
            Clear history to atime
            @param atime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM history_atime\
                         WHERE atime <= ?", (atime,))

    def get_from_atime(self, atime):
        """
            Get history ids from atime
            @param atime as int
            @return modified history ids as [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT DISTINCT history.rowid\
                                  FROM history, history_atime\
                                  WHERE history_atime.history_id=history.rowid\
                                  AND atime >= ?", (atime,))
            return list(itertools.chain(*result))

    def get_empties(self):
        """
            Get empties history entries (without atime)
            @return history ids as [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT history.rowid FROM history\
                                  WHERE NOT EXISTS (\
                                    SELECT rowid FROM history_atime AS ha\
                                    WHERE ha.history_id=history.rowid)")
            return list(itertools.chain(*result))

    def get(self, atime):
        """
            Get history for atime (current day)
            @param atime as int
            @return (str, str, int)
        """
        one_day = 86400
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT history.rowid, title, uri, atime\
                                  FROM history, history_atime\
                                  WHERE history.rowid=history_atime.history_id\
                                  AND atime >= ? AND atime <= ?\
                                  ORDER BY atime DESC",
                                 (atime, atime + one_day))
            return list(result)

    def get_id(self, uri):
        """
            Get history id
            @param uri as str
            @return history_id as int
        """
        with SqlCursor(self) as sql:
            uri = uri.rstrip('/')
            result = sql.execute("SELECT rowid\
                                  FROM history\
                                  WHERE uri=?",
                                 (uri,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_title(self, history_id):
        """
            Get history title
            @param history_id as int
            @return title as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT title\
                                  FROM history\
                                  WHERE rowid=?", (history_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_uri(self, history_id):
        """
            Get history uri
            @param history_id as int
            @return uri as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT uri\
                                  FROM history\
                                  WHERE rowid=?", (history_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_guid(self, history_id):
        """
            Get history item guid
            @param history_id as int
            @return guid as str
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT guid\
                                  FROM history\
                                  WHERE rowid=?", (history_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_mtime(self, history_id):
        """
            Get history mtime
            @param history_id as int
            @return mtime as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT mtime\
                                  FROM history\
                                  WHERE rowid=?", (history_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_atimes(self, history_id):
        """
            Get history access times
            @param history_id as int
            @return [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT atime\
                                  FROM history_atime\
                                  WHERE history_id=?", (history_id,))
            return list(itertools.chain(*result))

    def get_id_by_guid(self, guid):
        """
            Get id for guid
            @param guid as str
            @return id as int
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM history\
                                  WHERE guid=?", (guid,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_ids_for_mtime(self, mtime):
        """
            Get ids that need to be synced related to mtime
            @param mtime as int
            @return [int]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM history\
                                  WHERE mtime > ?", (mtime,))
            return list(itertools.chain(*result))

    def get_match(self, uri, ssl_force=False):
        """
            Try to get best uri matching
            @parma uri as str
            @param ssl_force as bool
            @return str
        """
        with SqlCursor(self) as sql:
            if ssl_force:
                filter = ("https://%{}%".format(uri),)
            else:
                filter = ("http%://%{}%".format(uri),)
            result = sql.execute("SELECT uri\
                                  FROM history\
                                  WHERE uri like ?\
                                  ORDER BY length(uri) ASC\
                                  LIMIT 1", filter)
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def set_title(self, history_id, title):
        """
            Set history title
            @param history_id as int
            @param title as str
            @param commit as bool
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE history\
                         SET title=?\
                         WHERE rowid=?", (title, history_id,))

    def get_populars(self, limit):
        """
            Get popular items in history
            @param limit as int
            @return [(id, title, uri, score)] as [(int, str, str, int)]
        """
        with SqlCursor(self) as sql:
            request = "SELECT rowid, title, uri FROM history\
                       ORDER BY popularity DESC LIMIT ?"
            result = sql.execute(request, (limit,))
            return list(result)
        return []

    def get_populars_by_netloc(self, netloc, limit):
        """
            Get popular items by netloc
            @param netloc as str
            @param limit as bool
            @return [(id, title, uri)]
        """
        with SqlCursor(self) as sql:
            if netloc:
                result = sql.execute("\
                                SELECT rowid,\
                                       uri,\
                                       uri,\
                                       title,\
                                       1\
                                FROM history\
                                WHERE netloc=?\
                                AND popularity!=0\
                                ORDER BY popularity DESC,\
                                mtime DESC\
                                LIMIT ?", (netloc, limit))
            else:
                result = sql.execute("\
                                SELECT rowid,\
                                       uri,\
                                       netloc,\
                                       netloc,\
                                       COUNT(uri)\
                                FROM history\
                                GROUP BY netloc\
                                ORDER BY MAX(popularity) DESC,\
                                mtime DESC\
                                LIMIT ?", (limit,))
            return list(result)

    def get_opened_pages(self):
        """
            Get page with opened state
            @return [(uri, title)]
        """
        with SqlCursor(self, True) as sql:
            try:
                result = sql.execute("SELECT uri, title\
                                      FROM history\
                                      WHERE opened=1")
                return list(result)
            finally:
                sql.execute("UPDATE history\
                             SET opened=0\
                             WHERE opened=1")

    def set_page_state(self, uri, mtime=None):
        """
            Mark page with uri as opened if mtime is not None
            @param uri as str
            @param mtime as double
        """
        if uri is None:
            return
        uri = uri.rstrip('/')
        with SqlCursor(self, True) as sql:
            if mtime is None:
                sql.execute("UPDATE history\
                             SET opened=0\
                             WHERE uri=?\
                             AND opened=1", (uri,))
            else:
                sql.execute("UPDATE history\
                             SET opened=1 WHERE uri=?\
                             AND mtime=?", (uri, mtime))

    def set_atimes(self, history_id, atimes):
        """
            Set history atime
            @param history_id as int
            @param atimes as [int]
            @param commit as bool
        """
        with SqlCursor(self, True) as sql:
            current_atimes = self.get_atimes(history_id)
            for atime in atimes:
                if atime not in current_atimes:
                    sql.execute("INSERT INTO history_atime (history_id, atime)\
                                 VALUES (?, ?)", (history_id, atime))

    def set_mtime(self, history_id, mtime):
        """
            Set history mtime
            @param history_id as int
            @param mtime as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("UPDATE history\
                         SET mtime=? where rowid=?", (mtime, history_id))

    def search(self, search, limit):
        """
            Search string in db (uri and title)
            @param search as str
            @param limit as int
            @return [(id, title, uri)] as [(int, str, str)]
        """
        words = search.lower().split()
        items = []
        with SqlCursor(self) as sql:
            filters = ()
            for word in words:
                filters += ("%" + word + "%", "%" + word + "%")
            filters += (limit,)
            request = "SELECT rowid, title, uri FROM history"
            if words:
                request += " WHERE"
                words_copy = list(words)
                while words_copy:
                    word = words_copy.pop(0)
                    request += " (title LIKE ? OR uri LIKE ?)"
                    if words_copy:
                        request += " AND "
            request += " ORDER BY length(uri) ASC LIMIT ?"
            try:
                result = sql.execute(request, filters)
                items = list(result)
            except:
                Logger.error("DatabaseHistory::search(): %s -> %s",
                             (request, filters))
        return items

    def reset_popularity(self, uri):
        """
            Reset popularity for uri
            @param uri as str
        """
        with SqlCursor(self, True) as sql:
            parsed = urlparse(uri)
            if parsed.scheme:
                sql.execute("UPDATE history SET popularity=0 WHERE uri=?",
                            (uri,))
            else:
                sql.execute("UPDATE history SET popularity=0 WHERE netloc=?",
                            (uri,))

    def exists_guid(self, guid):
        """
            Check if guid exists in db
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT guid FROM history\
                                  WHERE guid=?", (guid,))
            v = result.fetchone()
            return v is not None

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            c = sqlite3.connect(self.DB_PATH, 600.0)
            c.create_collation('LOCALIZED', LocalizedCollation())
            c.create_function("noaccents", 1, noaccents)
            return c
        except Exception as e:
            Logger.error("DatabaseHistory::get_cursor(): %s", e)
            exit(-1)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from locale import strcoll


class LocalizedCollation(object):
    """
        COLLATE LOCALIZED missing from default sqlite installation
        Android only
    """

    def __init__(self):
        pass

    def __call__(self, v1, v2):
        return strcoll(v1, v2)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio

from gettext import gettext as _
from time import mktime, time
from datetime import datetime
from locale import strcoll
from urllib.parse import urlparse

from eolie.helper_task import TaskHelper
from eolie.define import App, Type, TimeSpan, TimeSpanValues
from eolie.popover_uri_row import Row
from eolie.popover_uri_events import UriPopoverEvents
from eolie.popover_uri_content import UriPopoverContent
from eolie.popover_uri_suggestions import UriPopoverSuggestions
from eolie.popover_uri_webviews import UriPopoverWebviews
from eolie.popover_uri_input import Input


class UriPopover(Gtk.Popover, UriPopoverEvents, UriPopoverContent,
                 UriPopoverSuggestions, UriPopoverWebviews):
    """
        Show user bookmarks or search
    """

    def __init__(self, window):
        """
            Init popover
            @param window as Window
        """
        Gtk.Popover.__init__(self)
        UriPopoverEvents.__init__(self)
        UriPopoverContent.__init__(self)
        UriPopoverSuggestions.__init__(self)
        UriPopoverWebviews.__init__(self)
        self.__cancellable = Gio.Cancellable.new()
        self._task_helper = TaskHelper()
        self.set_modal(False)
        window.register(self, False)
        self._window = window
        self.get_style_context().add_class("box-shadow")
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/PopoverUri.ui")
        builder.connect_signals(self)
        self.__scrolled_bookmarks = builder.get_object("scrolled_bookmarks")
        self.__infobar = builder.get_object("infobar")
        self.__infobar_confirm = Gtk.Button()
        self.__infobar_confirm.show()
        self.__infobar_select = Gtk.ComboBoxText()
        self.__infobar_select.show()
        self.__infobar.get_content_area().add(self.__infobar_select)
        self.__infobar.add_action_widget(self.__infobar_confirm, 1)
        self.__infobar.add_button(_("Cancel"), 2)
        self._history_box = builder.get_object("history_box")
        self._history_box.bind_model(self._history_model,
                                     self.__on_item_create)
        self._search_box = builder.get_object("search_box")
        self._stack = builder.get_object("stack")
        self.__tags = builder.get_object("tags")
        self._tags_box = builder.get_object("tags_box")
        self._tags_box.set_sort_func(self.__sort_tags)
        self._remove_button = builder.get_object("remove_button")
        self._bookmarks_count = builder.get_object("count")
        self._bookmarks_box = builder.get_object("bookmarks_box")
        self._bookmarks_box.bind_model(self._bookmarks_model,
                                       self.__on_item_create)
        self._calendar = builder.get_object("calendar")
        self.add(builder.get_object("widget"))
        self.connect("map", self.__on_map)
        self.connect("unmap", self.__on_unmap)

    def popup(self, child):
        """
            Popup popover and wanted child
            @param child as str
        """
        Gtk.Popover.popup(self)
        # Add a new view for importing bookmarks
        if child == "bookmarks" and App().bookmarks.is_empty():
            grid = Gtk.Grid()
            image = Gtk.Image.new_from_icon_name("bookmark-new-symbolic",
                                                 Gtk.IconSize.BUTTON)
            grid.add(image)
            button = Gtk.Button.new_with_label(_("Import bookmarks"))
            button.connect("clicked", self._on_import_button_clicked)
            grid.add(button)
            grid.set_column_spacing(5)
            grid.set_property("valign", Gtk.Align.CENTER)
            grid.set_property("halign", Gtk.Align.CENTER)
            grid.show_all()
            self._stack.add(grid)
            self._stack.set_visible_child(grid)
        else:
            GLib.idle_add(self._stack.set_visible_child_name, child)

    def set_search_text(self, value):
        """
            Set search model
            @param value as str
        """
        self.__cancellable.cancel()
        self.__cancellable = Gio.Cancellable.new()
        for child in self._search_box.get_children():
            child.destroy()
        self.search_value(value, self.__cancellable)
        self.add_suggestions(value, self.__cancellable)
        self.add_webviews(value, self.__cancellable)
        self._stack.set_visible_child_name("search")

    @property
    def input(self):
        """
            Get input type
            @return Input
        """
        return self._input

#######################
# PROTECTED           #
#######################
    def _get_current_box(self):
        """
            Get current box
            @return Gtk.ListBox
        """
        name = self._stack.get_visible_child_name()
        box = None
        if name == "bookmarks":
            box = self._bookmarks_box
        elif name == "history":
            box = self._history_box
        elif name == "search":
            box = self._search_box
        return box

    def _get_current_input_box(self):
        """
            Get current box with input
            @return Gtk.ListBox
        """
        box = None
        if self._input == Input.SEARCH:
            box = self._search_box
        elif self._input == Input.TAGS:
            box = self._tags_box
        elif self._input == Input.BOOKMARKS:
            box = self._bookmarks_box
        return box

    def _on_import_button_clicked(self, button):
        """
            Sync with Firefox Sync
            @param button as Gtk.Button
        """
        self.hide()
        from eolie.dialog_import_bookmarks import ImportBookmarksDialog
        dialog = ImportBookmarksDialog(self._window)
        dialog.run()

    def _on_remove_button_clicked(self, button):
        """
            Remove bookmarks
            @param button as Gtk.Button
        """
        for row in self._bookmarks_box.get_selected_rows():
            item_id = row.item.get_property("id")
            if App().sync_worker is not None:
                guid = App().bookmarks.get_guid(item_id)
                App().sync_worker.remove_from_bookmarks(guid)
            App().bookmarks.remove(item_id)
            self._bookmarks_box.remove(row)
            self._remove_button.hide()
        App().bookmarks.clean_tags()

    def _on_selected_rows_changed(self, listbox):
        """
            Show delete button if needed
            @param listbox as Gtk.ListBox
        """
        self._remove_button.show()

    def _on_row_selected(self, listbox, row):
        """
            Scroll to row
            @param listbox as Gtk.ListBox
            @param row as Row
        """
        if row is None:
            return

        # Update titlebar
        if row.item.get_property("type") == Type.SUGGESTION:
            title = row.item.get_property("title")
            self._window.toolbar.title.entry.set_text_entry(title)
        else:
            uri = row.item.get_property("uri")
            if not uri:
                return
            parsed = urlparse(uri)
            if parsed.scheme not in ["http", "https", "file"]:
                uri = ""
            self._window.toolbar.title.entry.set_text_entry(uri)
        # Scroll to row
        scrolled = listbox.get_ancestor(Gtk.ScrolledWindow)
        if scrolled is None:
            return
        y = row.translate_coordinates(listbox, 0, 0)[1]
        adj = scrolled.get_vadjustment().get_value()
        if y + row.get_allocated_height() >\
                scrolled.get_allocated_height() + adj or\
                y - row.get_allocated_height() < 0 + adj:
            scrolled.get_vadjustment().set_value(y)

    def _on_search_map(self, widget):
        """
            Init search
            @param widget as Gtk.Widget
        """
        self._input = Input.NONE
        if not self._search_box.get_children():
            self.set_search_text("")

    def _on_history_map(self, widget):
        """
            Init search
            @param widget as Gtk.Widget
        """
        self._input = Input.NONE
        now = datetime.now()
        self._calendar.select_month(now.month - 1, now.year)
        self._calendar.select_day(now.day)

    def _on_close_map(self, widget):
        """
            Close popover
            @param widget as Gtk.Widget
        """
        self._window.close_popovers()

    def _on_bookmarks_map(self, widget):
        """
            Init bookmarks
            @param widget as Gtk.Widget/None
        """
        current = None
        row = self._tags_box.get_selected_row()
        if row is not None:
            current = row.item.get_property("id")
        self._input == Input.TAGS
        for child in self._tags_box.get_children():
            self._tags_box.remove(child)
        static = [(Type.POPULARS,
                   # Translators: Plural
                   _("Popular")),
                  (Type.RECENTS,
                   # Translators: Plural
                   _("Recent")),
                  (Type.UNCLASSIFIED,
                   _("Unclassified"))]
        GLib.idle_add(self._add_tags,
                      static + App().bookmarks.get_all_tags(), current)

    def _on_day_selected(self, calendar):
        """
            Show history for day
            @param calendar as Gtk.Calendar
        """
        (year, month, day) = calendar.get_date()
        date = datetime(year, month + 1, day, 0, 0)
        atime = mktime(date.timetuple())
        result = App().history.get(atime)
        self._history_model.remove_all()
        GLib.idle_add(self._add_history_items, result, (year, month, day))
        self.__infobar.hide()

    def _on_clear_history_clicked(self, button):
        """
            Ask user for confirmation
            @param button as Gtk.Button
        """
        self.__infobar_confirm.set_label(button.get_label())
        self.__infobar_select.remove_all()
        self.__infobar_select.append(TimeSpan.HOUR, _("From the past hour"))
        self.__infobar_select.append(TimeSpan.DAY, _("From the past day"))
        self.__infobar_select.append(TimeSpan.WEEK, _("From the past week"))
        self.__infobar_select.append(TimeSpan.FOUR_WEEK,
                                     _("From the past four weeks"))
        self.__infobar_select.append(TimeSpan.CUSTOM, _("From selected day"))
        self.__infobar_select.append(TimeSpan.FOREVER, _("From the beginning"))
        self.__infobar_select.set_active_id(TimeSpan.HOUR)
        self.__infobar.show()
        # GTK 3.20 https://bugzilla.gnome.org/show_bug.cgi?id=710888
        self.__infobar.queue_resize()

    def _on_infobar_response(self, infobar, response_id):
        """
            Handle user response and remove wanted history ids
            @param infobar as Gtk.InfoBar
            @param response_id as int
        """
        if response_id == 1:
            active_id = self.__infobar_select.get_active_id()
            if active_id == TimeSpan.CUSTOM:
                (year, month, day) = self._calendar.get_date()
                date = "%02d/%02d/%s" % (day, month + 1, year)
                atime = mktime(
                    datetime.strptime(date, "%d/%m/%Y").timetuple())
            else:
                atime = int(time() - TimeSpanValues[active_id] / 1000000)
            self._task_helper.run(self.__clear_history, atime)
        infobar.hide()

#######################
# PRIVATE             #
#######################
    def __clear_history(self, atime):
        """
            Clear history for wanted atime
            @param atime as double
            @thread safe
        """
        App().history.clear_from(atime)
        GLib.idle_add(self._on_day_selected, self._calendar)
        if App().sync_worker is not None:
            for history_id in App().history.get_empties():
                App().history.remove(history_id)

    def __sort_tags(self, row1, row2):
        """
            Sort tags
            @param row1 as Row
            @param row2 as Row
        """
        if row1.item.get_property("type") < 0:
            return False
        return strcoll(row1.item.get_property("title"),
                       row2.item.get_property("title"))

    def __on_map(self, widget):
        """
            Resize
            @param widget as Gtk.Widget
        """
        self._input = Input.NONE
        self._search_box.get_style_context().remove_class("kbd-input")
        self._bookmarks_box.get_style_context().remove_class("kbd-input")
        self._tags_box.get_style_context().remove_class("kbd-input")
        size = self._window.get_size()
        width = min(800, size[0])
        height = min(700, size[1] * 0.9)
        self.set_size_request(width, height)
        self.__scrolled_bookmarks.set_size_request(width * 0.4, -1)

    def __on_unmap(self, widget):
        """
            Switch to bookmarks
            @param widget as Gtk.Widget
        """
        self._stack.set_visible_child_name("bookmarks")
        self._bookmarks_model.remove_all()
        for child in self._tags_box.get_children():
            child.destroy()
        for child in self._search_box.get_children():
            child.destroy()

    def __on_row_edited(self, row):
        """
            Edit bookmark associated to row
            @param row as Row
        """
        from eolie.widget_bookmark_edit import BookmarkEditWidget
        widget = BookmarkEditWidget(row.item.get_property("id"))
        widget.show()
        self._stack.add(widget)
        self._stack.set_visible_child(widget)

    def __on_item_create(self, item):
        """
            Add child to box
            @param item as Item
        """
        child = Row(item, self._window)
        child.connect("edited", self.__on_row_edited)
        return child



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GLib, WebKit2

from eolie.widget_label_indicator import LabelIndicator
from eolie.utils import resize_favicon, update_popover_internals, emit_signal
from eolie.utils import get_round_surface, get_char_surface, get_safe_netloc
from eolie.define import App
from eolie.logger import Logger


class SitesManagerChild(Gtk.ListBoxRow):
    """
        Child showing snapshot, title and favicon
    """

    def __init__(self, netloc, window, is_ephemeral):
        """
            Init child
            @param netloc as str
            @param window as Window
            @param is_ephemeral as bool
        """
        Gtk.ListBoxRow.__init__(self)
        self.__window = window
        self.__netloc = netloc
        self.__minimal = None
        self.__current_child = None
        self.__is_ephemeral = is_ephemeral
        self.__webviews = []
        self.__connected_ids = []
        self.__scroll_timeout_id = None
        self.set_margin_top(1)
        self.get_style_context().add_class("sidebar-item")
        self.set_property("has-tooltip", True)
        self.connect("query-tooltip", self.__on_query_tooltip)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/SitesManagerChild.ui")
        builder.connect_signals(self)
        widget = builder.get_object("widget")
        self.__netloc_label = builder.get_object("netloc_label")
        self.__indicator_label = LabelIndicator(True)
        self.__indicator_label.set_hexpand(True)
        self.__indicator_label.show()
        builder.get_object("grid").attach(self.__indicator_label, 1, 0, 1, 1)
        self.__separator = builder.get_object("separator")
        self.__image = builder.get_object("image")
        self.add(widget)
        self.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, [],
                             Gdk.DragAction.MOVE)
        self.drag_source_add_text_targets()
        self.connect("drag-begin", self.__on_drag_begin)
        self.connect("drag-data-get", self.__on_drag_data_get)
        self.drag_dest_set(Gtk.DestDefaults.DROP | Gtk.DestDefaults.MOTION,
                           [], Gdk.DragAction.MOVE)
        self.drag_dest_add_text_targets()
        self.connect("drag-data-received", self.__on_drag_data_received)
        self.connect("drag-motion", self.__on_drag_motion)
        self.connect("drag-leave", self.__on_drag_leave)
        self.set_property("margin", 2)

    def add_webview(self, webview):
        """
            Add webview
            @param webview as WebView
            @param uri as str
        """
        if webview not in self.__webviews:
            self.__webviews.append(webview)
            webview.connect("load-changed", self.on_webview_load_changed)
            webview.connect("is-playing-audio",
                            self.__on_webview_is_playing_audio)
            webview.connect("notify::favicon",
                            self.__on_webview_favicon_changed)
            self.__indicator_label.add()
            self.__indicator_label.mark(webview)
            self.__netloc_label.set_text(get_safe_netloc(webview.uri))
            favicon_path = App().art.get_favicon_path(webview.uri)
            if favicon_path is not None:
                self.__image.set_from_file(favicon_path)

    def remove_webview(self, webview):
        """
            Remove webview and destroy self if no more webview
            @param webview as WebView
        """
        if webview in self.__webviews:
            self.__webviews.remove(webview)
            webview.disconnect_by_func(self.__on_webview_favicon_changed)
            webview.disconnect_by_func(self.on_webview_load_changed)
            webview.disconnect_by_func(self.__on_webview_is_playing_audio)
            self.__indicator_label.remove()
            self.__indicator_label.mark(webview)
            if self.__webviews:
                self.__set_favicon(self.__webviews[0])

    def show_label(self, show):
        """
            Show netloc label
            @param show as bool
        """
        if show:
            self.__netloc_label.show()
        else:
            self.__netloc_label.hide()

    def update_favicon(self):
        """
            Update favicon for current webview
        """
        if self.__window.container.webview in self.__webviews:
            self.__set_favicon(self.__window.container.webview)

    def reset(self, netloc):
        """
            Reset widget to new netloc
            @param netloc as str
        """
        if netloc != self.__netloc:
            self.__netloc = netloc

    def on_webview_load_changed(self, webview, event):
        """
            Update widget content
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event in [WebKit2.LoadEvent.STARTED,
                     WebKit2.LoadEvent.COMMITTED]:
            self.__image.set_from_icon_name(
                "network-transmit-receive-symbolic", Gtk.IconSize.MENU)
            self.__image.get_style_context().add_class("image-rotate")
        elif event == WebKit2.LoadEvent.FINISHED:
            self.__set_favicon(webview)

    @property
    def indicator_label(self):
        """
            Get indicator
            @return IndicatorLabel
        """
        return self.__indicator_label

    @property
    def empty(self):
        """
            True if no webview associated
            @return bool
        """
        return len(self.__webviews) == 0

    @property
    def webviews(self):
        """
            Get webviews
            @return [webview]
        """
        return self.__webviews

    @property
    def is_ephemeral(self):
        """
            True if ephemeral
            @return bool
        """
        return self.__is_ephemeral

    @property
    def netloc(self):
        """
            Get netloc
            @return str
        """
        return self.__netloc

#######################
# PROTECTED           #
#######################
    def _on_button_press_event(self, eventbox, event):
        """
            Hide popover or close webview
            @param eventbox as Gtk.EventBox
            @param event as Gdk.Event
        """
        try:
            if event.button == 2:
                for webview in self.__webviews:
                    self.__window.container.try_close_webview(webview)
                return True
            elif event.button == 3:
                from eolie.menu_sites import SitesMenu
                from eolie.menu_move_to import MoveToMenu
                sites_menu = SitesMenu(self.__webviews, self.__window)
                sites_menu.show()
                moveto_menu = MoveToMenu(self.__webviews, self.__window)
                moveto_menu.show()
                popover = Gtk.PopoverMenu.new()
                popover.add(sites_menu)
                popover.add(moveto_menu)
                popover.child_set_property(moveto_menu,
                                           "submenu", "moveto")
                popover.set_relative_to(eventbox)
                popover.set_position(Gtk.PositionType.RIGHT)
                popover.forall(update_popover_internals)
                popover.show()
                return True
        except Exception as e:
            Logger.error("SitesManagerChild::_on_button_press_event: %s", e)

#######################
# PRIVATE             #
#######################
    def __set_favicon(self, webview):
        """
            Set webview favicon
            @param webview as WebView
        """
        if webview.is_playing_audio():
            self.__image.set_from_icon_name("audio-speakers-symbolic",
                                            Gtk.IconSize.BUTTON)
        else:
            self.__image.get_style_context().remove_class("image-rotate")
            artwork = App().art.get_icon_theme_artwork(webview.uri,
                                                       webview.is_ephemeral)
            if artwork is not None:
                self.__image.set_from_icon_name(
                    artwork, Gtk.IconSize.LARGE_TOOLBAR)
            else:
                surface = webview.get_favicon()
                if surface is not None:
                    surface = get_round_surface(surface,
                                                webview.get_scale_factor(),
                                                surface.get_width() / 4)
                    self.__image.set_from_surface(resize_favicon(surface))
                else:
                    surface = get_char_surface(get_safe_netloc(webview.uri)[0])
                    self.__image.set_from_surface(surface)

    def __on_webview_is_playing_audio(self, webview, playing):
        """
            Update favicon
            @param webview as WebView
            @param playing as bool
        """
        self.__set_favicon(webview)

    def __on_webview_favicon_changed(self, webview, *ignore):
        """
            Set favicon
            @param webview as WebView
        """
        if self.__image.get_icon_name()[0] == "network-transmit-receive-symbolic":
            return
        if webview.get_favicon() is not None:
            self.__set_favicon(webview)

    def __on_query_tooltip(self, widget, x, y, keyboard, tooltip):
        """
            Show tooltip if needed
            @param widget as Gtk.Widget
            @param x as int
            @param y as int
            @param keyboard as bool
            @param tooltip as Gtk.Tooltip
        """
        tooltip = "<b>%s</b>" % GLib.markup_escape_text(self.__netloc)
        for webview in self.__webviews:
            title = webview.get_title()
            if title is not None:
                tooltip += "\n%s" % GLib.markup_escape_text(title)
        widget.set_tooltip_markup(tooltip)

    def __on_drag_begin(self, widget, context):
        """
            Set icon
            @param widget as Gtk.Widget
            @param context as Gdk.DragContext
        """
        surface = self.__image.get_property("surface")
        if surface is None:
            return
        pixbuf = Gdk.pixbuf_get_from_surface(surface,
                                             0, 0,
                                             surface.get_width(),
                                             surface.get_height())

        widget.drag_source_set_icon_pixbuf(pixbuf)
        del pixbuf

    def __on_drag_data_get(self, widget, context, data, info, time):
        """
            Send netloc
            @param widget as Gtk.Widget
            @param context as Gdk.DragContext
            @param data as Gtk.SelectionData
            @param info as int
            @param time as int
        """
        data.set_text(self.__netloc, len(self.__netloc))

    def __on_drag_data_received(self, widget, context, x, y, data, info, time):
        """
            Move track
            @param widget as Gtk.Widget
            @param context as Gdk.DragContext
            @param x as int
            @param y as int
            @param data as Gtk.SelectionData
            @param info as int
            @param time as int
        """
        height = self.get_allocated_height()
        if y > height / 2:
            up = False
        else:
            up = True
        try:
            netloc = data.get_text()
            emit_signal(self, "moved", netloc, up)
        except:
            pass

    def __on_drag_motion(self, widget, context, x, y, time):
        """
            Add style
            @param widget as Gtk.Widget
            @param context as Gdk.DragContext
            @param x as int
            @param y as int
            @param time as int
        """
        height = self.get_allocated_height()
        if y > height / 2:
            self.get_style_context().add_class("drag-up")
            self.get_style_context().remove_class("drag-down")
        else:
            self.get_style_context().remove_class("drag-up")
            self.get_style_context().add_class("drag-down")

    def __on_drag_leave(self, widget, context, time):
        """
            Remove style
            @param widget as Gtk.Widget
            @param context as Gdk.DragContext
            @param time as int
        """
        self.get_style_context().remove_class("drag-up")
        self.get_style_context().remove_class("drag-down")



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re

from eolie.css_rule_style import CSSStyleRule
from eolie.css_rule_media import CSSMediaRule
from eolie.css_rule_import import CSSImportRule
from eolie.css_rule_supports import CSSSupportsRule


class CSSRuleList:
    """
        Represent a list of rules
    """

    def __init__(self, css, uri, cancellable):
        """
            Init rule
            @param css as str
            @param uri as str
            @param cancellable as Gio.Cancellable
        """
        self.__uri = uri
        self.__rules = []
        for child in self.__get_children(css):
            child = child.strip()
            # Remove any comment
            child = re.sub(r"\/\*[^*]*[^/]*\*\/", "", child)
            if child.startswith("@media"):
                rule = CSSMediaRule(child, uri, cancellable)
            elif child.startswith("@supports"):
                rule = CSSSupportsRule(child, uri, cancellable)
            elif child.startswith("@import"):
                rule = CSSImportRule(child, uri, cancellable)
            elif child.startswith("@"):
                # Ignore others at-rules
                continue
            else:
                rule = CSSStyleRule(child)
            self.__rules.append(rule)

    @property
    def css_text(self):
        """
            Get css text for rules
        """
        css = [rule.css_text for rule in self.__rules]
        return "".join(css)

    @property
    def populated(self):
        """
            True if rule is populated
            @return bool
        """
        for rule in self.__rules:
            if not rule.populated:
                return False
        return True

#######################
# PRIVATE             #
#######################
    def __get_children(self, css):
        """
            Get children for css
            @param css as str
            @return [str]
        """
        children = []
        css = css.replace("\n", "").strip()
        css_split = css.split("}")
        bracket_count = 0
        children = []
        subcss = []
        while css_split:
            item = css_split.pop(0)
            for import_css in re.findall("@import[^;]*;", item):
                children.append(import_css)
                item.replace(import_css, "")
            bracket_count += item.count("{") - 1
            subcss.append(item)
            if bracket_count == 0:
                children.append("%s}" % "}".join(subcss))
                subcss = []
        return children



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, WebKit2, Gio

from urllib.parse import urlparse

from eolie.define import App, LoadingState
from eolie.helper_task import TaskHelper
from eolie.utils import get_snapshot, resize_favicon, emit_signal
from eolie.utils import get_round_surface, get_char_surface, get_safe_netloc


class WebViewArtwork:
    """
        Handle webview artwork: snapshot and favicon
    """

    def __init__(self):
        """
            Init class
        """
        self.__helper = TaskHelper()
        self.__cancellable = Gio.Cancellable()
        self.__surface = None
        self.__snapshot_id = None
        self.__scroll_event_id = None
        self.__is_snapshot_valid = False
        self.connect("uri-changed", self.__on_uri_changed)
        self.connect("scroll-event", self.__on_scroll_event)
        self.connect("notify::favicon", self.__on_webview_favicon_changed)

    @property
    def surface(self):
        """
            Get webview snapshot surface
            @return cairo.Surface
        """
        return self.__surface

    @property
    def is_snapshot_valid(self):
        """
            True if snapshot is valid for current URI
            @return bool
        """
        return self.__is_snapshot_valid

#######################
# PROTECTED           #
#######################
    def _on_load_changed(self, webview, event):
        """
            Update sidebar/urlbar
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event == WebKit2.LoadEvent.STARTED:
            self._loading = True
            self.__is_snapshot_valid = False
            self.__cancellable.cancel()
            self.__cancellable = Gio.Cancellable()
            if self.__snapshot_id is not None:
                GLib.source_remove(self.__snapshot_id)
                self.__snapshot_id = None
        elif event == WebKit2.LoadEvent.FINISHED:
            if self.__snapshot_id is not None:
                GLib.source_remove(self.__snapshot_id)
                self.__snapshot_id = None
            if self._loading:
                self.__snapshot_id = GLib.timeout_add(2500,
                                                      self.__set_snapshot)
            self._loading = False

#######################
# PRIVATE             #
#######################
    def __on_webview_favicon_changed(self, webview, *ignore):
        """
            Set favicon based on current webview favicon
            @param webview as WebView
        """
        parsed = urlparse(self.uri)
        if self.is_ephemeral or parsed.scheme not in ["http", "https"]:
            return
        surface = self.get_favicon()
        if surface is not None:
            surface = get_round_surface(surface,
                                        self.get_scale_factor(),
                                        surface.get_width() / 4)
            self.__set_favicon_from_surface(surface, self.uri)
        else:
            self.__set_favicon_from_surface(None, self.uri)

    def __set_snapshot(self):
        """
            Set webpage preview
        """
        self.__snapshot_id = None
        # Only save page if bookmarked
        save = False
        for uri in [self.uri, self.loaded_uri]:
            if uri is not None and App().bookmarks.get_id(uri) is not None:
                save = True
                break
        self.get_snapshot(WebKit2.SnapshotRegion.VISIBLE,
                          WebKit2.SnapshotOptions.NONE,
                          self.__cancellable,
                          get_snapshot,
                          self.__on_snapshot,
                          save)

    def __set_favicon_from_surface(self, surface, uri):
        """
            Set favicon for surface
            @param favicon_db as WebKit2.FaviconDatabase
            @param result as Gio.AsyncResult
            @param uri as str
        """
        # Get a default favicon
        if surface is None:
            surface = get_char_surface(get_safe_netloc(uri)[0])
        # Save webview favicon
        if surface is not None:
            resized = resize_favicon(surface)
            self.__save_favicon_to_cache(resized,
                                         uri)
            if self.loaded_uri is not None and uri != self.loaded_uri:
                self.__save_favicon_to_cache(resized,
                                             self.loaded_uri)

    def __save_favicon_to_cache(self, surface, uri):
        """
            Save favicon to cache
            @param surface as cairo.Surface
            @param uri as str
        """
        # Save favicon for URI
        self.__helper.run(App().art.save_artwork,
                          uri,
                          surface,
                          "favicon")

    def __on_uri_changed(self, webview, param):
        """
            Handle JS updates
            @param webview as WebKit2.WebView
            @param param as GObject.ParamSpec
        """
        if webview.get_uri() is not None and\
                not webview.is_loading() and\
                not webview.is_ephemeral:
            if self.__snapshot_id is not None:
                GLib.source_remove(self.__snapshot_id)
            self.__snapshot_id = GLib.timeout_add(2500,
                                                  self.__set_snapshot)

    def __on_snapshot(self, surface, save):
        """
            Cache snapshot
            @param surface as cairo.Surface
            @param uri as str
            @param save as bool
        """
        self.__surface = surface
        self.__is_snapshot_valid = True
        emit_signal(self, "snapshot-changed", surface)
        if not save or self.loading_state in [LoadingState.ERROR,
                                              LoadingState.STOPPED]:
            return
        if App().settings.get_value("night-mode"):
            prefix = "start_dark"
        else:
            prefix = "start_light"
        App().art.save_artwork(self.uri, surface, prefix)
        if self.loaded_uri is not None and self.uri != self.loaded_uri:
            App().art.save_artwork(self.loaded_uri, surface, prefix)

    def __on_scroll_event(self, widget, event):
        """
            Update snapshot
            @param widget as WebView
            @param event as Gdk.EventScroll
        """
        def update_snapshot():
            self.__scroll_event_id = None
            self.get_snapshot(WebKit2.SnapshotRegion.VISIBLE,
                              WebKit2.SnapshotOptions.NONE,
                              self.__cancellable,
                              get_snapshot,
                              self.__on_snapshot,
                              False)
        if self.__snapshot_id is not None:
            GLib.source_remove(self.__snapshot_id)
            self.__snapshot_id = None
        if self.__scroll_event_id is not None:
            GLib.source_remove(self.__scroll_event_id)
        self.__scroll_event_id = GLib.timeout_add(1000, update_snapshot)



# Copyright (c) 2014-2019 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps

from eolie.define import App
from eolie.logger import Logger
# For lint
App()


def signals(f):
    """
        Decorator to init signal helper
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        ret = f(*args, **kwargs)
        SignalsHelper.__init__(args[0])
        args[0].init(ret)

    return wrapper


def signals_map(f):
    """
        Decorator to init signal helper
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        ret = f(*args, **kwargs)
        SignalsHelper.__init__(args[0])
        args[0].init_map(ret)

    return wrapper


class SignalsHelper():
    """
        Helper for autoconnect/disconnect signals on map
        Keep signals cached when unmapped
    """

    def __init__(self):
        """
            Init helper
        """
        if not hasattr(self, "_connected"):
            self._connected = {}
            self.__signal_ids = []
            self.__allow_duplicate = []
            self.__cached = {}

    def allow_duplicate(self, callback_str):
        """
            Allow duplicate signal emit for callback while hidden
            @param callback_str as str
        """
        callback = getattr(self, callback_str)
        if callback not in self.__allow_duplicate:
            self.__allow_duplicate.append(callback)

    def init(self, signals):
        """
            Init signals
            @param signals as []
        """
        self._connect_signals(signals, False)
        self.__init(signals, False)

    def init_map(self, signals):
        """
            Init signals on map
            @param signals as []
        """
        self.__signal_ids.append(
                     self.connect("map",
                                  lambda x: self._connect_signals(
                                    signals, True)))
        self.__init(signals, True)

#######################
# PROTECTED           #
#######################
    def _connect_signals(self, signals, on_map):
        """
            Connect signals
            @param signals as []
            @param on_map on bool
        """
        for (obj, signal, callback_str) in signals:
            if obj is None:
                Logger.debug("Can't connect signal: %s -> %s",
                             signal, callback_str)
                continue
            name = "%s_%s" % (obj, signal)
            if name in self._connected.keys():
                continue
            callback = getattr(self, callback_str)
            if on_map:
                self._connected[name] = obj.connect(signal,
                                                    self.__on_signal,
                                                    callback)
            else:
                self._connected[name] = obj.connect(signal, callback)

    def _disconnect_signals(self, signals):
        """
            Disconnect signals
            @param signals as []
        """
        for (obj, signal, callback_str) in signals:
            if obj is None:
                Logger.warning("Can't disconnect signal: %s -> %s",
                               signal, callback_str)
                continue
            name = "%s_%s" % (obj, signal)
            if name not in self._connected.keys():
                continue
            connect_id = self._connected[name]
            obj.disconnect(connect_id)
            del self._connected[name]

#######################
# PRIVATE             #
#######################
    def __init(self, signals, on_map):
        """
            Init signals
            @param signals as []
            @param on_map on bool
        """
        if on_map:
            self.connect("map", self.__on_map)
            self.connect("unmap", self.__on_unmap)
        self.connect("destroy",
                     lambda x: self._disconnect_signals(signals))

    def __on_map(self, widget):
        """
            Run cached callbacks
            @param widget as Gtk.Widget
        """
        for callback in self.__cached.keys():
            (obj, callback_args) = self.__cached[callback]
            callback(obj, *callback_args)
        self.__cached = {}

    def __on_unmap(self, widget):
        """
            Disconnect initial map signal
            @param widget as Gtk.Widget
        """
        for signal_id in self.__signal_ids:
            self.disconnect(signal_id)
        self.__signal_ids = []

    def __on_signal(self, obj, *args):
        """
            Keep callback in cache if self not mapped
        """
        callback = args[-1]
        callback_args = args[:-1]
        if self.get_mapped() or callback in self.__allow_duplicate:
            callback(obj, *callback_args)
        else:
            self.__cached[callback] = (obj, callback_args)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

from gettext import gettext as _

from eolie.define import App


class WebViewJsSignals:
    """
        Handle webview scripts signals
    """

    def __init__(self):
        """
            Init class
        """
        self.__js_blocker_count = 0
        self.__js_blocker_timeout_id = None
        self.connect("script-dialog", self.__on_script_dialog)

#######################
# PROTECTED           #
#######################

#######################
# PRIVATE             #
#######################
    def __reset_js_blocker(self):
        """
            Reset js blocker
        """
        self.__js_blocker_count = 0
        self.__js_blocker_timeout_id = None

    def __on_script_dialog(self, webview, dialog):
        """
            Show message to user
            @param webview as WebView
            @param dialog as WebKit2.ScriptDialog
        """
        if self.__js_blocker_timeout_id is not None:
            GLib.source_remove(self.__js_blocker_timeout_id)
            self.__js_blocker_timeout_id = None
        message = dialog.get_message()
        if message.startswith("@EOLIE_CSS_URI@"):
            webview.load_css_uri(message)
        elif message.startswith("@EOLIE_CSS_TEXT@"):
            webview.load_css_text(message)
        # Credentials message
        elif message.startswith("@EOLIE_SUBMIT@"):
            webview.add_credentials(message)
        # Input menu message
        elif message.startswith("@EOLIE_FORM_MENU_MESSAGE@"):
            webview.show_form_menu(message)
        # Focus message
        elif message.startswith("@EOLIE_FOCUS_MESSAGE@"):
            self.window.toolbar.title.show_input_warning(self)
        # OpenSearch message
        elif message.startswith("@EOLIE_OPENSEARCH@"):
            uri = message.replace("@EOLIE_OPENSEARCH@", "")
            App().search.install_engine(uri, self.window)
        # Populars view message
        elif message.startswith("@EOLIE_HIDE_BOOKMARK_POPULARS@"):
            uri = message.replace("@EOLIE_HIDE_BOOKMARK_POPULARS@", "")
            App().bookmarks.reset_popularity(uri)
        # Populars view message
        elif message.startswith("@EOLIE_HIDE_HISTORY_POPULARS@"):
            uri = message.replace("@EOLIE_HIDE_HISTORY_POPULARS@", "")
            App().history.reset_popularity(uri)
        # Here we handle JS flood
        elif self.__js_blocker_count > 5:
            self.__js_blocker_count = 0
            self.window.toolbar.title.show_message(
                _("Eolie is going to close this page because it is broken"))
            self.window.container.close_webview(self.view)
        # Webpage message
        else:
            self.window.toolbar.title.show_javascript(dialog)
            self.__js_blocker_count += 1
            self.__js_blocker_timeout_id = GLib.timeout_add(
                1000,
                self.__reset_js_blocker)
        return True



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from urllib.parse import urlparse


class CredentialsPopover(Gtk.Popover):
    """
        Tell user to save form credentials
    """

    def __init__(self, window):
        """
            Init popover
            @param window as Window
        """
        Gtk.Popover.__init__(self)
        self.__webview = window.container.webview
        self.set_modal(False)
        window.register(self)
        builder = Gtk.Builder()
        builder.add_from_resource('/org/gnome/Eolie/PopoverCredentials.ui')
        builder.connect_signals(self)
        self.__label = builder.get_object('label')
        parsed = urlparse(self.__webview.credentials_uri)
        builder.get_object('uri').set_text(parsed.netloc)
        self.add(builder.get_object('widget'))

#######################
# PROTECTED           #
#######################
    def _on_cancel_button_clicked(self, button):
        """
            Close popover
            @param button as Gtk.Button
        """
        self.popdown()

    def _on_save_button_clicked(self, button):
        """
            Save user_form_name and pass_form_name
            @param button as Gtk.Button
        """
        self.__webview.save_credentials()
        self.destroy()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Fork of https://github.com/firefox-services/syncclient
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, GObject

from pickle import dump, load
from hashlib import sha256
import json
from fcntl import flock, LOCK_EX, LOCK_NB, LOCK_UN
from time import time, sleep

from eolie.helper_task import TaskHelper
from eolie.define import App, EOLIE_DATA_PATH
from eolie.sqlcursor import SqlCursor
from eolie.helper_passwords import PasswordsHelper
from eolie.logger import Logger
from eolie.utils import emit_signal


TOKENSERVER_URL = "https://%s/" %\
    App().settings.get_value("firefox-sync-server").get_string()


class SyncWorker(GObject.Object):
    """
       Manage sync with firefox server, will start syncing on init
    """

    __gsignals__ = {
        "syncing": (GObject.SignalFlags.RUN_FIRST, None, (str,))
    }

    def check_modules():
        """
            True if deps are installed
            @return bool
        """
        from importlib import util
        fxa = util.find_spec("fxa")
        crypto = util.find_spec("Crypto")
        hawk = util.find_spec("requests_hawk")
        if fxa is None:
            Logger.info("PyFxA missing: pip3 install --user pyfxa")
        if crypto is None:
            Logger.info("Cyrpto missing: pip3 install --user pycrypto")
        if hawk is None:
            Logger.info("Hawk missing: pip3 install --user requests_hawk")
        return fxa is not None and crypto is not None and hawk is not None

    def __init__(self):
        """
            Init worker
        """
        GObject.Object.__init__(self)
        self.__sync_cancellable = Gio.Cancellable()
        self.__timeout_id = None
        self.__username = ""
        self.__password = ""
        self.__token = ""
        self.__uid = ""
        self.__keyB = b""
        self.__mtimes = {"bookmarks": 0.1, "history": 0.1, "passwords": 0.1}
        # We do not create this here because it's slow down Eolie startup
        # See __firefox_sync property
        self.__mz = None
        self.__state_lock = True
        self.__session = None
        self.__syncing = False
        self.__syncing_pendings = False
        try:
            self.__pending_records = load(open(EOLIE_DATA_PATH +
                                               "/firefox_sync_pendings.bin",
                                               "rb"))
            self.__sync_pendings()
        except:
            self.__pending_records = {"history": [],
                                      "passwords": [],
                                      "bookmarks": []}
        self.__helper = PasswordsHelper()
        if App().settings.get_value("enable-firefox-sync"):
            self.set_credentials()

    def login(self, attributes, password, code):
        """
            Login to service
            @param attributes as {}
            @param password as str
            @param code as str
        """
        self.__username = ""
        self.__password = ""
        self.__uid = ""
        self.__token = ""
        self.__keyB = b""
        if attributes is None or not attributes["login"] or not password:
            Logger.warning("SyncWorker::login(): %s", attributes)
            return
        from base64 import b64encode
        import json
        session = None
        self.__username = attributes["login"]
        self.__password = password
        # Connect to firefox sync
        try:
            session = self.__firefox_sync.login(
                self.__username, password, code)
            bid_assertion, key = self.__firefox_sync.\
                get_browserid_assertion(session)
            self.__token = session.token
            self.__uid = session.uid
            self.__keyB = session.keys[1]
            keyB_encoded = b64encode(self.__keyB).decode("utf-8")
            record = {"uid": self.__uid,
                      "token": self.__token,
                      "keyB": keyB_encoded}
            self.__helper.clear_sync(self.__helper.store_sync,
                                     self.__username,
                                     json.dumps(record))
        except Exception as e:
            self.__helper.clear_sync(self.__helper.store_sync,
                                     self.__username,
                                     "")
            raise(e)

    def new_session(self):
        """
            Start a new session
        """
        # Just reset session, will be set by get_session_bulk_keys()
        self.__session = None

    def set_credentials(self):
        """
            Set credentials using Secret
        """
        self.__helper.get_sync(self.__set_credentials)

    def pull_loop(self):
        """
            Start pulling every hours
        """
        def loop():
            self.pull()
            return True
        if App().settings.get_value("enable-firefox-sync"):
            self.pull()
            if self.__timeout_id is None:
                self.__timeout_id = GLib.timeout_add_seconds(600, loop)

    def pull(self, force=False):
        """
            Start pulling from Firefox sync
            @param force as bool
        """
        if Gio.NetworkMonitor.get_default().get_network_available() and\
                self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__pull, force)

    def push(self):
        """
            Start pushing to Firefox sync
            Will push all data
        """
        if Gio.NetworkMonitor.get_default().get_network_available() and\
                self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__push,)

    def push_history(self, history_id):
        """
            Push history id
            @param history_id as int
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__push_history, history_id)

    def push_bookmark(self, bookmark_id):
        """
            Push bookmark id
            @param bookmark_id as int
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__push_bookmark, bookmark_id)

    def push_password(self, user_form_name, user_form_value, pass_form_name,
                      pass_form_value, uri, form_uri, uuid):
        """
            Push password
            @param user_form_name as str
            @param user_form_value as str
            @param pass_form_name as str
            @param pass_form_value as str
            @param uri as str
            @param form_uri as str
            @param uuid as str
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__push_password,
                            user_form_name, user_form_value, pass_form_name,
                            pass_form_value, uri, form_uri, uuid)

    def remove_from_history(self, guid):
        """
            Remove history guid from remote history
            @param guid as str
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__remove_from_history, guid)

    def remove_from_bookmarks(self, guid):
        """
            Remove bookmark guid from remote bookmarks
            @param guid as str
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__remove_from_bookmarks, guid)

    def remove_from_passwords(self, uuid):
        """
            Remove password from passwords collection
            @param uuid as str
        """
        if self.__username:
            task_helper = TaskHelper()
            task_helper.run(self.__remove_from_passwords, uuid)

    def delete_secret(self):
        """
            Delete sync secret
        """
        self.__username = ""
        self.__password = ""
        self.__session = None
        self.__helper.clear_sync(None)

    def stop(self, force=False):
        """
            Stop update, if force, kill session too
            @param force as bool
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None
        self.__sync_cancellable.cancel()
        self.__sync_cancellable = Gio.Cancellable()
        if force:
            self.__session = None

    def save_pendings(self):
        """
            Save pending records
        """
        try:
            dump(self.__pending_records,
                 open(EOLIE_DATA_PATH + "/firefox_sync_pendings.bin", "wb"))
        except Exception as e:
            Logger.Error("SyncWorker::save_pendings(): %s", e)

    @property
    def syncing(self):
        """
            True if syncing
            @return bool
        """
        return self.__syncing

    @property
    def status(self):
        """
            True if sync is working
            @return bool
        """
        try:
            if self.__username:
                self.__get_session_bulk_keys()
                self.__firefox_sync.client.info_collections()
                return True
        except Exception as e:
            Logger.error("SyncWorker::status(): %s", e)
        return False

    @property
    def username(self):
        """
            Get username
            @return str
        """
        return self.__username

#######################
# PRIVATE             #
#######################
    @property
    def __firefox_sync(self):
        """
            Get firefox sync, create if None
        """
        if self.__mz is None:
            self.__mz = FirefoxSync()
        return self.__mz

    def __get_session_bulk_keys(self):
        """
            Get session decrypt keys
            @return keys as (b"", b"")
        """
        if self.__session is None:
            from fxa.core import Session as FxASession
            from fxa.crypto import quick_stretch_password
            self.__session = FxASession(self.__firefox_sync.fxa_client,
                                        self.__username,
                                        quick_stretch_password(
                                            self.__username,
                                            self.__password),
                                        self.__uid,
                                        self.__token)
            self.__session.keys = [b"", self.__keyB]
        self.__session.check_session_status()
        bid_assertion, key = self.__firefox_sync.get_browserid_assertion(
            self.__session)
        bulk_keys = self.__firefox_sync.connect(bid_assertion, key)
        return bulk_keys

    def __update_state(self):
        """
            Update state file
        """
        try:
            f = open(EOLIE_DATA_PATH + "/firefox_sync.bin", "wb")
            # Lock file
            flock(f, LOCK_EX | LOCK_NB)
            self.__mtimes = self.__firefox_sync.client.info_collections()
            dump(self.__mtimes, f)
            # Unlock file
            flock(f, LOCK_UN)
        except Exception as e:
            # Not an error, just the lock exception
            Logger.info("SyncWorker::__update_state(): %s", e)

    def __sync_pendings(self):
        """
            Sync pendings record
        """
        try:
            if Gio.NetworkMonitor.get_default().get_network_available() and\
                    self.__username and not self.__syncing_pendings:
                self.__syncing_pendings = True
                Logger.sync_debug("Elements to push to Firefox sync: %s",
                                  self.__pending_records)
                self.__check_worker()
                bulk_keys = self.__get_session_bulk_keys()
                for key in self.__pending_records.keys():
                    while self.__pending_records[key]:
                        try:
                            record = self.__pending_records[key].pop(0)
                            Logger.sync_debug("syncing %s", record)
                            if "bmkUri" in record.keys():
                                emit_signal(self, "syncing",
                                            record["bmkUri"])
                            elif "histUri" in record.keys():
                                emit_signal(self, "syncing",
                                            record["histUri"])
                            elif "hostname" in record.keys():
                                emit_signal(self, "syncing",
                                            record["hostname"])
                            self.__firefox_sync.add(record, key, bulk_keys)
                        except:
                            self.__pending_records[key].append(record)
                self.__update_state()
                self.__syncing_pendings = False
        except Exception as e:
            Logger.error("SyncWorker::__sync_pendings(): %s", e)
            self.__syncing_pendings = False

    def __push_history(self, history_id, sync=True):
        """
            Push history item
            @param history is as int
            @param sync as bool
        """
        try:
            record = {}
            atimes = App().history.get_atimes(history_id)
            guid = App().history.get_guid(history_id)
            record["histUri"] = App().history.get_uri(history_id)
            record["id"] = guid
            record["title"] = App().history.get_title(history_id)
            record["visits"] = []
            for atime in atimes:
                record["visits"].append({"date": atime * 1000000,
                                         "type": 1})
            self.__pending_records["history"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.error("SyncWorker::__push_history(): %s", e)

    def __push_bookmark(self, bookmark_id, sync=True):
        """
            Push bookmark
            @param bookmark_id as int
            @param sync as bool
        """
        try:
            parent_guid = App().bookmarks.get_parent_guid(bookmark_id)
            # No parent, move it to unfiled
            if parent_guid is None:
                parent_guid = "unfiled"
            parent_id = App().bookmarks.get_id_by_guid(parent_guid)
            record = {}
            record["bmkUri"] = App().bookmarks.get_uri(bookmark_id)
            record["id"] = App().bookmarks.get_guid(bookmark_id)
            record["title"] = App().bookmarks.get_title(bookmark_id)
            record["tags"] = App().bookmarks.get_tags(bookmark_id)
            record["parentid"] = parent_guid
            record["parentName"] = App().bookmarks.get_parent_name(bookmark_id)
            record["type"] = "bookmark"
            self.__pending_records["bookmarks"].append(record)
            parent_guid = App().bookmarks.get_guid(parent_id)
            parent_name = App().bookmarks.get_title(parent_id)
            children = App().bookmarks.get_children(parent_guid)
            record = {}
            record["id"] = parent_guid
            record["type"] = "folder"
            # A parent with parent as unfiled needs to be moved to places
            # Firefox internal
            grand_parent_guid = App().bookmarks.get_parent_guid(parent_id)
            if grand_parent_guid == "unfiled":
                grand_parent_guid = "places"
            record["parentid"] = grand_parent_guid
            record["parentName"] = App().bookmarks.get_parent_name(parent_id)
            record["title"] = parent_name
            record["children"] = children
            self.__pending_records["bookmarks"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.error("SyncWorker::__push_bookmark(): %s", e)

    def __push_password(self, user_form_name, user_form_value, pass_form_name,
                        pass_form_value, uri, form_uri, uuid, sync=True):
        """
            Push password
            @param user_form_name as str
            @param user_form_value as str
            @param pass_form_name as str
            @param pass_form_value as str
            @param uri as str
            @param uuid as str
            @param sync as bool
        """
        try:
            record = {}
            record["id"] = "{%s}" % uuid
            record["hostname"] = uri
            record["formSubmitURL"] = form_uri
            record["httpRealm"] = None
            record["username"] = user_form_value
            record["password"] = pass_form_value
            record["usernameField"] = user_form_name
            record["passwordField"] = pass_form_name
            mtime = int(time() * 1000)
            record["timeCreated"] = mtime
            record["timePasswordChanged"] = mtime
            self.__pending_records["passwords"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.error("SyncWorker::__push_password(): %s", e)

    def __remove_from_history(self, guid, sync=True):
        """
            Remove from history
            @param guid as str
            @param sync as bool
        """
        try:
            record = {}
            record["id"] = guid
            record["type"] = "item"
            record["deleted"] = True
            self.__pending_records["history"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.sync_debug("SyncWorker::__remove_from_history(): %s", e)

    def __remove_from_bookmarks(self, guid, sync=True):
        """
            Remove from history
            @param guid as str
            @param sync as bool
        """
        try:
            record = {}
            record["id"] = guid
            record["type"] = "bookmark"
            record["deleted"] = True
            self.__pending_records["bookmarks"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.sync_debug("SyncWorker::__remove_from_bookmarks(): %s", e)

    def __remove_from_passwords(self, uuid, sync=True):
        """
            Remove password from passwords collection
            @param uuid as str
            @param sync as bool
        """
        try:
            record = {}
            record["id"] = uuid
            record["deleted"] = True
            self.__pending_records["passwords"].append(record)
            if sync:
                self.__sync_pendings()
        except Exception as e:
            Logger.sync_debug("SyncWorker::__remove_from_passwords(): %s", e)

    def __pull(self, force):
        """
            Pull bookmarks, history, ... from Firefox Sync
            @param force as bool
        """
        if self.__syncing:
            return
        Logger.sync_debug("Start pulling")
        self.__syncing = True
        self.__sync_cancellable.cancel()
        self.__sync_cancellable = Gio.Cancellable()
        try:
            if force:
                raise
            self.__mtimes = load(open(EOLIE_DATA_PATH + "/firefox_sync.bin",
                                      "rb"))
        except:
            self.__mtimes = {"bookmarks": 0.1,
                             "history": 0.1,
                             "passwords": 0.1}
        try:
            self.__check_worker()

            bulk_keys = self.__get_session_bulk_keys()
            new_mtimes = self.__firefox_sync.client.info_collections()

            self.__check_worker()
            ########################
            # Passwords Management #
            ########################
            try:
                Logger.sync_debug("local passwords: %s, remote passwords: %s",
                                  self.__mtimes["passwords"],
                                  new_mtimes["passwords"])
                # Only pull if something new available
                if self.__mtimes["passwords"] != new_mtimes["passwords"]:
                    self.__pull_passwords(bulk_keys)
            except:
                pass  # No passwords in sync

            self.__check_worker()
            ######################
            # History Management #
            ######################
            try:
                Logger.sync_debug("local history: %s, remote history: %s",
                                  self.__mtimes["history"],
                                  new_mtimes["history"])
                # Only pull if something new available
                if self.__mtimes["history"] != new_mtimes["history"]:
                    self.__pull_history(bulk_keys)
            except:
                pass

            self.__check_worker()
            ########################
            # Bookmarks Management #
            ########################
            try:
                Logger.sync_debug("local bookmarks: %s, remote bookmarks: %s",
                                  self.__mtimes["bookmarks"],
                                  new_mtimes["bookmarks"])
                # Only pull if something new available
                if self.__mtimes["bookmarks"] != new_mtimes["bookmarks"]:
                    self.__pull_bookmarks(bulk_keys)
            except:
                pass
            self.__update_state()
            Logger.sync_debug("Stop pulling")
        except Exception as e:
            Logger.error("SyncWorker::__pull(): %s", e)
        self.__syncing = False

    def __push(self):
        """
            Push bookmarks, history, ... to Firefox Sync
            @param force as bool
        """
        if self.__syncing:
            return
        Logger.sync_debug("Start pushing")
        self.__syncing = True
        self.__sync_cancellable.cancel()
        self.__sync_cancellable = Gio.Cancellable()
        try:
            self.__check_worker()

            ########################
            # Passwords Management #
            ########################
            self.__helper.get_all(self.__on_helper_get_all)

            self.__check_worker()
            ######################
            # History Management #
            ######################
            for history_id in App().history.get_from_atime(0):
                self.__push_history(history_id, False)

            self.__check_worker()
            ########################
            # Bookmarks Management #
            ########################
            for (bookmark_id, title, uri) in App().bookmarks.get_bookmarks():
                self.__push_bookmark(bookmark_id, False)
            self.__check_worker()
            self.__sync_pendings()
            Logger.sync_debug("Stop pushing")
        except Exception as e:
            Logger.error("SyncWorker::__push(): %s", e)
        self.__syncing = False

    def __pull_bookmarks(self, bulk_keys):
        """
            Pull from bookmarks
            @param bulk_keys as KeyBundle
            @raise StopIteration
        """
        Logger.sync_debug("pull bookmarks")
        SqlCursor.add(App().bookmarks)
        records = self.__firefox_sync.get_records("bookmarks", bulk_keys)
        children_array = []
        for record in records:
            self.__check_worker()
            if record["modified"] < self.__mtimes["bookmarks"]:
                continue
            sleep(0.01)
            bookmark = record["payload"]
            bookmark_id = App().bookmarks.get_id_by_guid(bookmark["id"])
            # Nothing to apply, continue
            if App().bookmarks.get_mtime(bookmark_id) >= record["modified"]:
                continue
            Logger.sync_debug("pulling %s", record)
            # Deleted bookmark
            if "deleted" in bookmark.keys():
                App().bookmarks.remove(bookmark_id)
            # Keep folder only for firefox compatiblity
            elif "type" in bookmark.keys() and bookmark["type"] == "folder"\
                    and bookmark["id"] is not None\
                    and bookmark["title"]:
                if bookmark_id is None:
                    bookmark_id = App().bookmarks.add(bookmark["title"],
                                                      bookmark["id"],
                                                      bookmark["id"],
                                                      [],
                                                      0)
                # Will calculate position later
                if "children" in bookmark.keys():
                    children_array.append(bookmark["children"])
            # We have a bookmark, add it
            elif "type" in bookmark.keys() and bookmark["type"] == "bookmark"\
                    and bookmark["id"] is not None\
                    and bookmark["title"]:
                # Update bookmark
                if bookmark_id is not None:
                    App().bookmarks.set_title(bookmark_id,
                                              bookmark["title"])
                    App().bookmarks.set_uri(bookmark_id,
                                            bookmark["bmkUri"])
                    # Update tags
                    current_tags = App().bookmarks.get_tags(bookmark_id)
                    for tag in App().bookmarks.get_tags(bookmark_id):
                        if "tags" in bookmark.keys() and\
                                tag not in bookmark["tags"]:
                            tag_id = App().bookmarks.get_tag_id(tag)
                            current_tags.remove(tag)
                            App().bookmarks.del_tag_from(tag_id,
                                                         bookmark_id)
                    if "tags" in bookmark.keys():
                        for tag in bookmark["tags"]:
                            # Tag already associated
                            if tag in current_tags:
                                continue
                            tag_id = App().bookmarks.get_tag_id(tag)
                            if tag_id is None:
                                tag_id = App().bookmarks.add_tag(tag)
                            App().bookmarks.add_tag_to(tag_id,
                                                       bookmark_id)
                # Add a new bookmark
                else:
                    bookmark_id = App().bookmarks.get_id(bookmark["bmkUri"])
                    # Add a new bookmark
                    if bookmark_id is None:
                        # Use parent name if no bookmarks tags
                        if "tags" not in bookmark.keys() or\
                                not bookmark["tags"]:
                            if "parentName" in bookmark.keys() and\
                                    bookmark["parentName"]:
                                bookmark["tags"] = [bookmark["parentName"]]
                            else:
                                bookmark["tags"] = []
                        bookmark_id = App().bookmarks.add(bookmark["title"],
                                                          bookmark["bmkUri"],
                                                          bookmark["id"],
                                                          bookmark["tags"],
                                                          0)
                    else:
                        # Update guid
                        App().bookmarks.set_guid(bookmark_id, bookmark["id"])
            # Update parent name if available
            if bookmark_id is not None and "parentName" in bookmark.keys():
                App().bookmarks.set_parent(bookmark_id,
                                           bookmark["parentid"],
                                           bookmark["parentName"])
            App().bookmarks.set_mtime(bookmark_id,
                                      record["modified"])
        # Update bookmark position
        for children in children_array:
            position = 0
            for child in children:
                bid = App().bookmarks.get_id_by_guid(child)
                App().bookmarks.set_position(bid,
                                             position)
                position += 1
        App().bookmarks.clean_tags()  # Will commit
        SqlCursor.remove(App().bookmarks)

    def __pull_passwords(self, bulk_keys):
        """
            Pull from passwords
            @param bulk_keys as KeyBundle
            @raise StopIteration
        """
        Logger.sync_debug("pull passwords")
        records = self.__firefox_sync.get_records("passwords", bulk_keys)
        for record in records:
            self.__check_worker()
            if record["modified"] < self.__mtimes["passwords"]:
                continue
            sleep(0.01)
            Logger.sync_debug("pulling %s", record)
            password = record["payload"]
            password_id = password["id"].strip("{}")
            if "formSubmitURL" in password.keys():
                self.__helper.clear(password_id,
                                    self.__helper.store,
                                    password["usernameField"],
                                    password["username"],
                                    password["passwordField"],
                                    password["password"],
                                    password["hostname"],
                                    password["formSubmitURL"],
                                    password_id,
                                    None)
            elif "deleted" in password.keys():  # We assume True
                self.__helper.clear(password_id)

    def __pull_history(self, bulk_keys):
        """
            Pull from history
            @param bulk_keys as KeyBundle
            @raise StopIteration
        """
        Logger.sync_debug("pull history")
        records = self.__firefox_sync.get_records("history", bulk_keys)
        for record in records:
            self.__check_worker()
            if record["modified"] < self.__mtimes["history"]:
                continue
            sleep(0.01)
            history = record["payload"]
            keys = history.keys()
            history_id = App().history.get_id_by_guid(history["id"])
            # Check we have a valid history item
            if "histUri" in keys and\
                    "title" in keys and\
                    history["title"] and\
                    App().history.get_mtime(history_id) < record["modified"]:
                # Try to get visit date
                atimes = []
                try:
                    for visit in history["visits"]:
                        atimes.append(round(int(visit["date"]) / 1000000, 2))
                except:
                    continue
                Logger.sync_debug("pulling %s", record)
                title = history["title"].rstrip().lstrip()
                history_id = App().history.add(title,
                                               history["histUri"],
                                               record["modified"],
                                               history["id"],
                                               atimes,
                                               True)
            elif "deleted" in keys:
                history_id = App().history.get_id_by_guid(history_id)
                App().history.remove(history_id)

    def __set_credentials(self, attributes, password, uri, index, count):
        """
            Set credentials
            @param attributes as {}
            @param password as str
            @param uri as None
            @param index as int
            @param count as int
        """
        if attributes is None:
            return
        from base64 import b64decode
        import json
        try:
            self.__username = attributes["login"]
            record = json.loads(password)
            self.__token = record["token"]
            self.__uid = record["uid"]
            self.__keyB = b64decode(record["keyB"])
        except Exception as e:
            Logger.error("SyncWorker::__set_credentials(): %s" % e)

    def __check_worker(self):
        """
            Raise an exception if worker should not be syncing: error&cancel
        """
        if self.__sync_cancellable.is_cancelled():
            raise StopIteration("SyncWorker: cancelled")
        elif not self.__username:
            raise StopIteration("SyncWorker: missing username")
        elif not self.__token:
            raise StopIteration("SyncWorker: missing token")

    def __on_helper_get_all(self, attributes, password, uri, index, count):
        """
            Push password
            @param attributes as {}
            @param password as str
            @param uri as None
            @param index as int
            @param count as int
        """
        if attributes is None:
            return
        try:
            self.__check_worker()
            user_form_name = attributes["userform"]
            user_form_value = attributes["login"]
            pass_form_name = attributes["passform"]
            pass_form_value = password
            uri = attributes["hostname"]
            form_uri = attributes["formSubmitURL"]
            uuid = attributes["uuid"]
            task_helper = TaskHelper()
            task_helper.run(self.__push_password,
                            user_form_name, user_form_value, pass_form_name,
                            pass_form_value, uri, form_uri, uuid, False)
        except Exception as e:
            Logger.error("SyncWorker::__on_helper_get_all(): %s" % e)


class FirefoxSync(object):
    """
        Sync client
    """

    def __init__(self):
        """
            Init client
        """
        from fxa.core import Client as FxAClient
        self.__fxa_client = FxAClient()

    def login(self, login, password, code):
        """
            Login to FxA and get the keys.
            @param login as str
            @param password as str
            @param code as str
            @return fxaSession
        """
        fxaSession = self.__fxa_client.login(login, password, keys=True)
        if code:
            fxaSession.totp_verify(code)
        fxaSession.fetch_keys()
        return fxaSession

    def connect(self, bid_assertion, key):
        """
            Connect to sync using FxA browserid assertion
            @param session as fxaSession
            @return bundle keys as KeyBundle
        """
        state = None
        if key is not None:
            from binascii import hexlify
            state = hexlify(sha256(key).digest()[0:16])
        self.__client = SyncClient(bid_assertion, state)
        sync_keys = KeyBundle.fromMasterKey(
            key,
            "identity.mozilla.com/picl/v1/oldsync")

        # Fetch the sync bundle keys out of storage.
        # They're encrypted with the account-level key.
        keys = self.__decrypt_payload(self.__client.get_record("crypto",
                                                               "keys"),
                                      sync_keys)

        # There's some provision for using separate
        # key bundles for separate collections
        # but I haven't bothered digging through
        # to see what that's about because
        # it doesn't seem to be in use, at least on my account.
        if keys["collections"]:
            Logger.error("""no support for per-collection
                            key bundles yet sorry :-(""")
            return None

        # Now use those keys to decrypt the records of interest.
        from base64 import b64decode
        bulk_keys = KeyBundle(b64decode(keys["default"][0]),
                              b64decode(keys["default"][1]))
        return bulk_keys

    def get_records(self, collection, bulk_keys):
        """
            Return records payload
            @param collection as str
            @param bulk keys as KeyBundle
            @return [{}]
        """
        records = []
        for record in self.__client.get_records(collection):
            record["payload"] = self.__decrypt_payload(record, bulk_keys)
            records.append(record)
        return records

    def add(self, item, collection, bulk_keys):
        """
            Add bookmark
            @param bookmark as {}
            @param collection as str
            @param bulk_keys as KeyBundle
        """
        payload = self.__encrypt_payload(item, bulk_keys)
        record = {}
        record["modified"] = round(time(), 2)
        record["payload"] = payload
        record["id"] = item["id"]
        self.__client.put_record(collection, record)

    def get_browserid_assertion(self, session,
                                tokenserver_url=TOKENSERVER_URL):
        """
            Get browser id assertion and state
            @param session as fxaSession
            @return (bid_assertion, state) as (str, str)
        """
        bid_assertion = session.get_identity_assertion(tokenserver_url)
        return bid_assertion, session.keys[1]

    @property
    def client(self):
        """
            Get client
        """
        return self.__client

    @property
    def fxa_client(self):
        """
            Get fxa client
        """
        return self.__fxa_client

#######################
# PRIVATE             #
#######################
    def __encrypt_payload(self, record, key_bundle):
        """
            Encrypt payload
            @param record as {}
            @param key bundle as KeyBundle
            @return encrypted record payload
        """
        from Crypto.Cipher import AES
        from Crypto import Random
        from hmac import new
        from base64 import b64encode
        plaintext = json.dumps(record).encode("utf-8")
        # Input strings must be a multiple of 16 in length
        length = 16 - (len(plaintext) % 16)
        plaintext += bytes([length]) * length
        iv = Random.new().read(16)
        aes = AES.new(key_bundle.encryption_key, AES.MODE_CBC, iv)
        ciphertext = b64encode(aes.encrypt(plaintext))
        _hmac = new(key_bundle.hmac_key, ciphertext, sha256).hexdigest()
        payload = {"ciphertext": ciphertext.decode("utf-8"),
                   "IV": b64encode(iv).decode("utf-8"), "hmac": _hmac}
        return json.dumps(payload)

    def __decrypt_payload(self, record, key_bundle):
        """
            Descrypt payload
            @param record as str (json)
            @param key bundle as KeyBundle
            @return uncrypted record payload
        """
        from Crypto.Cipher import AES
        from hmac import new
        from base64 import b64decode
        j = json.loads(record["payload"])
        # Always check the hmac before decrypting anything.
        expected_hmac = new(key_bundle.hmac_key,
                            j['ciphertext'].encode("utf-8"),
                            sha256).hexdigest()
        if j['hmac'] != expected_hmac:
            raise ValueError("HMAC mismatch: %s != %s" % (j['hmac'],
                                                          expected_hmac))
        ciphertext = b64decode(j['ciphertext'])
        iv = b64decode(j['IV'])
        aes = AES.new(key_bundle.encryption_key, AES.MODE_CBC, iv)
        plaintext = aes.decrypt(ciphertext).strip().decode("utf-8")
        # Remove any CBC block padding,
        # assuming it's a well-formed JSON payload.
        plaintext = plaintext[:plaintext.rfind("}") + 1]
        return json.loads(plaintext)


class KeyBundle:
    """
        RFC-5869
    """

    def __init__(self, encryption_key, hmac_key):
        self.encryption_key = encryption_key
        self.hmac_key = hmac_key

    @classmethod
    def fromMasterKey(cls, master_key, info):
        key_material = KeyBundle.HKDF(master_key, None, info, 2 * 32)
        return cls(key_material[:32], key_material[32:])

    def HKDF_extract(salt, IKM, hashmod=sha256):
        """
            Extract a pseudorandom key suitable for use with HKDF_expand
            @param salt as str
            @param IKM as str
        """
        from hmac import new
        if salt is None:
            salt = b"\x00" * hashmod().digest_size
        return new(salt, IKM, hashmod).digest()

    def HKDF_expand(PRK, info, length, hashmod=sha256):
        """
            Expand pseudo random key and info
            @param PRK as str
            @param info as str
            @param length as int
        """
        from hmac import new
        from math import ceil
        digest_size = hashmod().digest_size
        N = int(ceil(length * 1.0 / digest_size))
        assert N <= 255
        T = b""
        output = []
        for i in range(1, N + 1):
            data = T + (info + chr(i)).encode()
            T = new(PRK, data, hashmod).digest()
            output.append(T)
        return b"".join(output)[:length]

    def HKDF(secret, salt, info, length, hashmod=sha256):
        """
            HKDF-extract-and-expand as a single function.
            @param secret as str
            @param salt as str
            @param info as str
            @param length as int
        """
        PRK = KeyBundle.HKDF_extract(salt, secret, hashmod)
        return KeyBundle.HKDF_expand(PRK, info, length, hashmod)


class TokenserverClient(object):
    """
        Client for the Firefox Sync Token Server.
    """

    def __init__(self, bid_assertion, client_state,
                 server_url=TOKENSERVER_URL):
        """
            Init client
            @param bid assertion as str
            @param client_state as ???
            @param server_url as str
        """
        self.__bid_assertion = bid_assertion
        self.__client_state = client_state
        self.__server_url = server_url

    def get_hawk_credentials(self, duration=None):
        """
            Asks for new temporary token given a BrowserID assertion
            @param duration as str
        """
        from requests import get
        authorization = 'BrowserID %s' % self.__bid_assertion
        headers = {
            'Authorization': authorization,
            'X-Client-State': self.__client_state
        }
        params = {}

        if duration is not None:
            params['duration'] = int(duration)

        url = self.__server_url.rstrip('/') + '/1.0/sync/1.5'
        raw_resp = get(url, headers=headers, params=params, verify=True)
        raw_resp.raise_for_status()
        return raw_resp.json()


class SyncClient(object):
    """
        Client for the Firefox Sync server.
    """

    def __init__(self, bid_assertion=None, client_state=None,
                 credentials={}, tokenserver_url=TOKENSERVER_URL):
        """
            Init client
            @param bid assertion as str
            @param client_state as ???
            @param credentials as {}
            @param server_url as str
        """
        from requests_hawk import HawkAuth
        if bid_assertion is not None and client_state is not None:
            ts_client = TokenserverClient(bid_assertion, client_state,
                                          tokenserver_url)
            credentials = ts_client.get_hawk_credentials()
        self.__user_id = credentials['uid']
        self.__api_endpoint = credentials['api_endpoint']
        self.__auth = HawkAuth(algorithm=credentials['hashalg'],
                               id=credentials['id'],
                               key=credentials['key'],
                               always_hash_content=False)

    def _request(self, method, url, **kwargs):
        """
            Utility to request an endpoint with the correct authentication
            setup, raises on errors and returns the JSON.
            @param method as str
            @param url as str
            @param kwargs as requests.request named args
        """
        from requests import request, exceptions
        url = self.__api_endpoint.rstrip('/') + '/' + url.lstrip('/')
        raw_resp = request(method, url, auth=self.__auth, **kwargs)
        raw_resp.raise_for_status()

        if raw_resp.status_code == 304:
            http_error_msg = '%s Client Error: %s for url: %s' % (
                raw_resp.status_code,
                raw_resp.reason,
                raw_resp.url)
            raise exceptions.HTTPError(http_error_msg, response=raw_resp)
        return raw_resp.json()

    def info_collections(self, **kwargs):
        """
            Returns an object mapping collection names associated with the
            account to the last-modified time for each collection.

            The server may allow requests to this endpoint to be authenticated
            with an expired token, so that clients can check for server-side
            changes before fetching an updated token from the Token Server.
        """
        return self._request('get', '/info/collections', **kwargs)

    def info_quota(self, **kwargs):
        """
            Returns a two-item list giving the user's current usage and quota
            (in KB). The second item will be null if the server
            does not enforce quotas.

            Note that usage numbers may be approximate.
        """
        return self._request('get', '/info/quota', **kwargs)

    def get_collection_usage(self, **kwargs):
        """
            Returns an object mapping collection names associated with the
            account to the data volume used for each collection (in KB).

            Note that these results may be very expensive as it calculates more
            detailed and accurate usage information than the info_quota method.
        """
        return self._request('get', '/info/collection_usage', **kwargs)

    def get_collection_counts(self, **kwargs):
        """
            Returns an object mapping collection names associated with the
            account to the total number of items in each collection.
        """
        return self._request('get', '/info/collection_counts', **kwargs)

    def delete_all_records(self, **kwargs):
        """
            Deletes all records for the user
        """
        return self._request('delete', '/', **kwargs)

    def get_records(self, collection, full=True, ids=None, newer=None,
                    limit=None, offset=None, sort=None, **kwargs):
        """
            Returns a list of the BSOs contained in a collection. For example:

            >>> ["GXS58IDC_12", "GXS58IDC_13", "GXS58IDC_15"]

            By default only the BSO ids are returned, but full objects can be
            requested using the full parameter. If the collection does not
            exist, an empty list is returned.

            :param ids:
                a comma-separated list of ids. Only objects whose id is in
                this list will be returned. A maximum of 100 ids may be
                provided.

            :param newer:
                a timestamp. Only objects whose last-modified time is strictly
                greater than this value will be returned.

            :param full:
                any value. If provided then the response will be a list of full
                BSO objects rather than a list of ids.

            :param limit:
                a positive integer. At most that many objects will be returned.
                If more than that many objects matched the query,
                an X-Weave-Next-Offset header will be returned.

            :param offset:
                a string, as returned in the X-Weave-Next-Offset header of a
                previous request using the limit parameter.

            :param sort:
                sorts the output:
                "newest" - orders by last-modified time, largest first
                "index" - orders by the sortindex, highest weight first
                "oldest" - orders by last-modified time, oldest first
        """
        params = kwargs.pop('params', {})
        if full:
            params['full'] = True
        if ids is not None:
            params['ids'] = ','.join(map(str, ids))
        if newer is not None:
            params['newer'] = newer
        if limit is not None:
            params['limit'] = limit
        if offset is not None:
            params['offset'] = offset
        if sort is not None and sort in ('newest', 'index', 'oldest'):
            params['sort'] = sort

        return self._request('get', '/storage/%s' % collection.lower(),
                             params=params, **kwargs)

    def get_record(self, collection, record_id, **kwargs):
        """Returns the BSO in the collection corresponding to the requested id.
        """
        return self._request('get', '/storage/%s/%s' % (collection.lower(),
                                                        record_id), **kwargs)

    def delete_record(self, collection, record_id, **kwargs):
        """Deletes the BSO at the given location.
        """
        try:
            return self._request('delete', '/storage/%s/%s' % (
                collection.lower(), record_id), **kwargs)
        except Exception as e:
            Logger.error("SyncClient::delete_record(): %s", e)

    def put_record(self, collection, record, **kwargs):
        """
            Creates or updates a specific BSO within a collection.
            The passed record must be a python object containing new data for
            the BSO.

            If the target BSO already exists then it will be updated with the
            data from the request body. Fields that are not provided will not
            be overwritten, so it is possible to e.g. update the ttl field of a
            BSO without re-submitting its payload. Fields that are explicitly
            set to null in the request body will be set to their default value
            by the server.

            If the target BSO does not exist, then fields that are not provided
            in the python object will be set to their default value
            by the server.

            Successful responses will return the new last-modified time for the
            collection.

            Note that the server may impose a limit on the amount of data
            submitted for storage in a single BSO.
        """
        import six
        # XXX: Workaround until request-hawk supports the json parameter. (#17)
        if isinstance(record, six.string_types):
            record = json.loads(record)
        record = record.copy()
        record_id = record.pop('id')
        headers = {}
        if 'headers' in kwargs:
            headers = kwargs.pop('headers')
        headers['Content-Type'] = 'application/json; charset=utf-8'

        return self._request('put', '/storage/%s/%s' % (
            collection.lower(), record_id), data=json.dumps(record),
            headers=headers, **kwargs)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, WebKit2

from eolie.widget_label_indicator import LabelIndicator
from eolie.define import ArtSize, MARGIN_SMALL, MARGIN
from eolie.utils import on_query_tooltip, update_popover_internals
from eolie.helper_signals import SignalsHelper, signals
from eolie.helper_gestures import GesturesHelper


class PagesManagerGenericChild(SignalsHelper, GesturesHelper):
    """
        Widget showing a webpage snapshot and title
    """

    @signals
    def __init__(self, webview, window):
        """
            Init child
            @param webview as WebView
            @param window as Window
        """
        GesturesHelper.__init__(self, self)

        self.__webview = webview
        self.__window = window

        self.get_style_context().add_class("sidebar-item")

        close_button = Gtk.Button.new_from_icon_name(
            "window-close-symbolic", Gtk.IconSize.LARGE_TOOLBAR)
        close_button.show()
        close_button.set_relief(Gtk.ReliefStyle.NONE)
        close_button.get_style_context().add_class("no-border-button")
        close_button.set_property("valign", Gtk.Align.CENTER)
        close_button.set_property("halign", Gtk.Align.END)
        close_button.connect("clicked", self.__on_close_button_clicked)

        self.__indicator_label = LabelIndicator(False)
        self.__indicator_label.show()
        self.__indicator_label.mark(webview)
        self.__indicator_label.set_hexpand(True)
        self.__indicator_label.set_margin_right(4)
        self.__indicator_label.set_property("halign", Gtk.Align.CENTER)
        self.__indicator_label.set_property("valign", Gtk.Align.CENTER)
        self.__indicator_label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__indicator_label.set_property("has-tooltip", True)
        self.__indicator_label.connect("query-tooltip", on_query_tooltip)
        if webview.title:
            self.__indicator_label.set_text(webview.title)

        self.__indicator_image = Gtk.Image.new()
        self.__indicator_image.set_property("halign", Gtk.Align.CENTER)
        self.__indicator_image.set_property("valign", Gtk.Align.CENTER)
        self.__indicator_image.get_style_context().add_class(
            "sidebar-item-image-overlay")

        self.__background_image = Gtk.Image.new_from_surface(webview.surface)
        self.__background_image.show()
        self.__background_image.get_style_context().add_class(
            "sidebar-item-image")

        grid = Gtk.Grid()
        grid.show()
        grid.get_style_context().add_class("sidebar-item-title")
        grid.set_column_spacing(MARGIN_SMALL)
        grid.set_vexpand(True)
        grid.set_property("valign", Gtk.Align.END)
        grid.set_property("margin", MARGIN_SMALL)
        grid.add(self.__indicator_label)
        grid.add(close_button)

        overlay = Gtk.Overlay.new()
        overlay.show()
        overlay.add(self.__background_image)
        overlay.add_overlay(grid)
        overlay.add_overlay(self.__indicator_image)

        self.set_property("halign", Gtk.Align.START)
        self.set_property("margin", MARGIN)
        # padding + border
        self.set_size_request(ArtSize.START_WIDTH + 8,
                              ArtSize.START_HEIGHT + 8)
        self.add(overlay)

        return [
            (webview, "snapshot-changed", "_on_webview_snapshot_changed"),
            (webview, "is-playing-audio", "_on_webview_is_playing_audio"),
            (webview, "title-changed", "_on_webview_title_changed"),
            (webview, "load-changed", "_on_webview_load_changed"),
            (webview, "destroy", "_on_webview_destroyed")
        ]

    @property
    def indicator_label(self):
        """
            Get indicator
            @return IndicatorLabel
        """
        return self.__indicator_label

    @property
    def webview(self):
        """
            Get linked webview
            @return WebView
        """
        return self.__webview

#######################
# PROTECTED           #
#######################
    def _on_secondary_press_gesture(self, x, y):
        """
            Show row menu
            @param x as int
            @param y as int
        """
        self.__window.container.try_close_webview(self.__webview)

    def _on_tertiary_press_gesture(self, eventbox, event):
        """
            Hide popover or close view
            @param eventbox as Gtk.EventBox
            @param event as Gdk.Event
        """
        from eolie.menu_move_to import MoveToMenu
        moveto_menu = MoveToMenu([self.__webview], self.__window, False)
        moveto_menu.show()
        popover = Gtk.PopoverMenu.new()
        popover.set_relative_to(eventbox)
        popover.set_position(Gtk.PositionType.BOTTOM)
        popover.add(moveto_menu)
        popover.forall(update_popover_internals)
        popover.show()

    def _on_webview_is_playing_audio(self, webview, playing):
        """
            Update favicon
            @param webview as WebView
            @param playing as bool
        """
        if webview.is_playing_audio():
            self.__indicator_image.set_from_icon_name(
                "audio-speakers-symbolic", Gtk.IconSize.DND)
            self.__indicator_image.show()
        else:
            self.__indicator_image.hide()
            self.__indicator_image.set_from_surface(None)

    def _on_webview_title_changed(self, webview, title):
        """
            Update title
            @param webview as WebView
            @param title as str
        """
        if title:
            self.__indicator_label.set_text(title)

    def _on_webview_load_changed(self, webview, event):
        """
            Update widget content
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if event != WebKit2.LoadEvent.FINISHED:
            self.__background_image.set_from_surface(None)
            self.__indicator_image.set_from_surface(None)
            self.__indicator_image.set_from_icon_name(
                "network-transmit-receive-symbolic", Gtk.IconSize.DND)
            self.__indicator_image.get_style_context().add_class(
                "image-rotate")
            self.__indicator_image.show()
        else:
            self.__indicator_image.set_from_surface(None)
            self.__indicator_image.get_style_context().remove_class(
                "image-rotate")
            self.__indicator_image.hide()

    def _on_webview_snapshot_changed(self, webview, surface):
        """
            Update preview with surface
            @param webview as WebView
            @param surface as cairo.surface
        """
        if webview.is_ephemeral:
            self.__background_image.set_from_icon_name(
                "user-not-tracked-symbolic",
                Gtk.IconSize.DIALOG)
        else:
            self.__background_image.set_from_surface(surface)

    def _on_webview_destroyed(self, webview):
        """
            Destroy self
            @param webview as WebView
        """
        self.destroy()

#######################
# PRIVATE             #
#######################
    def __on_close_button_clicked(self, button):
        """
            Destroy self
            @param button as Gtk.Button
        """
        self.__window.container.try_close_webview(self.__webview)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re

from eolie.logger import Logger


class CSSSupportsRule:
    """
        Represent a supports rule
    """

    def __init__(self, css, uri, cancellable):
        """
            Init rule
            @param css as str
            @param uri as str
            @param cancellable as Gio.Cancellable
        """
        self.__rules = None
        self.__condition = None
        try:
            # Get condition @supprt
            search = re.search('@supports ([^{]*){(.*)', css)
            if search is not None:
                from eolie.css_rule_list import CSSRuleList
                self.__condition = search.group(1)
                self.__rules = CSSRuleList(search.group(2), uri, cancellable)
        except Exception as e:
            Logger.error("CSSMediaRule::__init__(): %s -> %s", e, css)

    @property
    def css_text(self):
        """
            Get css text for rules
            @return str
        """
        if self.__rules is not None:
            css_text = self.__rules.css_text
            if css_text != "":
                return "@supports %s { %s } " % (self.__condition, css_text)
        return ""

    @property
    def populated(self):
        """
            True if rule is populated
            @return bool
        """
        if self.__rules is None:
            return True
        else:
            return self.__rules.populated

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import WebKit2, Gio, GLib

from urllib.parse import urlparse
from time import time

from eolie.define import App, LoadingType, LoadingState
from eolie.context import Context
from eolie.webview_errors import WebViewErrors
from eolie.webview_navigation import WebViewNavigation
from eolie.webview_signals import WebViewSignals
from eolie.webview_artwork import WebViewArtwork
from eolie.webview_state import WebViewState
from eolie.webview_credentials import WebViewCredentials
from eolie.webview_helpers import WebViewHelpers
from eolie.webview_night_mode import WebViewNightMode
from eolie.list import LinkedList
from eolie.utils import emit_signal
from eolie.logger import Logger


class WebView(WebKit2.WebView):
    """
        WebKit view
    """

    def new(window):
        """
            New webview
            @param window as Window
        """
        context = WebKit2.WebContext(
            process_swap_on_cross_site_navigation_enabled=True)
        Context(context)

        webview = WebKit2.WebView.new_with_context(context)
        webview.__class__ = WebViewMeta
        webview.__init(None, window)
        return webview

    def new_ephemeral(window):
        """
            New ephemeral webview
            @param window as Window
        """
        context = WebKit2.WebContext.new_ephemeral()
        Context(context)
        webview = WebKit2.WebView.new_with_context(context)
        webview.__class__ = WebViewMeta
        webview.__init(None, window)
        return webview

    def new_with_related_view(related, window):
        """
            Create a new WebView related to view
            @param related as WebView
            @param window as Window
            @return WebView
        """
        webview = WebKit2.WebView.new_with_related_view(related)
        webview.__class__ = WebViewMeta
        webview.__init(related, window)
        return webview

    def set_setting(self, key, value):
        """
            Set setting to value
            @param key as str
            @param value as GLib.Variant
        """
        settings = self.get_settings()
        if key == 'use-system-fonts':
            self.__set_system_fonts(settings)
        else:
            settings.set_property(key, value)

    def reload_bypass_cache(self):
        """
            Reload page, destroy night mode cache
        """
        self.remove_night_mode_cache()
        WebKit2.WebView.reload_bypass_cache(self)

    def update_zoom_level(self):
        """
            Update zoom level
        """
        try:
            if self.__related is not None:
                zoom_level = self.__related.window.zoom_level
            else:
                zoom_level = self.window.zoom_level
                _zoom_level = App().websettings.get("zoom", self.uri)
                if _zoom_level is not None:
                    zoom_level = _zoom_level / 100
        except Exception as e:
            Logger.error("WebView::update_zoom_level(): %s", e)
        Logger.debug("Update zoom level: %s", zoom_level)
        self.set_zoom_level(zoom_level)

    def is_playing_audio(self):
        """
            Do not return True if sound disabled
        """
        value = App().websettings.get("audio", self.uri)
        if value:
            return WebKit2.WebView.is_playing_audio(self)
        else:
            return False

    def update_sound_policy(self):
        """
            Update sound policy
        """
        value = App().websettings.get("audio", self.uri)
        self.__window.container.webview.set_is_muted(not value)
        if value:
            emit_signal(self, "is-playing-audio", self.is_playing_audio())
        else:
            emit_signal(self, "is-playing-audio", False)

    def print(self):
        """
            Show print dialog for current page
        """
        p = WebKit2.PrintOperation.new(self)
        p.run_dialog()

    def zoom_in(self):
        """
            Zoom in view
            @return current zoom after zoom in
        """
        parsed = urlparse(self.uri)
        if parsed.netloc in ["http", "https"]:
            current = App().websettings.get("zoom", self.uri)
        else:
            current = int(self.get_zoom_level() * 100)
        if current is None:
            current = int(self.window.zoom_level * 100)
        current += 10
        App().websettings.set("zoom", self.uri, current)
        self.set_zoom_level(current / 100)
        return current

    def zoom_out(self):
        """
            Zoom out view
            @return current zoom after zoom out
        """
        parsed = urlparse(self.uri)
        if parsed.netloc in ["http", "https"]:
            current = App().websettings.get("zoom", self.uri)
        else:
            current = int(self.get_zoom_level() * 100)
        if current is None:
            current = int(self.window.zoom_level * 100)
        current -= 10
        if current == 0:
            return 10
        App().websettings.set("zoom", self.uri, current)
        self.set_zoom_level(current / 100)
        return current

    def zoom_default(self):
        """
            Reset zoom level
            @return current zoom after zoom out
        """
        current = int(self.window.zoom_level * 100)
        App().websettings.set("zoom", self.uri, None)
        self.set_zoom_level(current / 100)
        return current

    def update_spell_checking(self, uri):
        """
            Update spell checking
        """
        context = self.get_context()
        codes = App().websettings.get_languages(uri)
        # If None, default user language
        if codes is None:
            locales = GLib.get_language_names()
            try:
                codes = [locales[0].split(".")[0]]
            except:
                codes = None
        if context.get_spell_checking_languages() != codes:
            context.set_spell_checking_languages(codes)

    def add_text_entry(self, text):
        """
            Add an uri to text entry list
            @param text as str
        """
        if text and text != self.__text_entry.value:
            item = LinkedList(text, None, self.__text_entry)
            self.__text_entry = item

    def clear_text_entry(self):
        """
            Clear text entry history
        """
        self.__text_entry = LinkedList("", None, None)

    def get_current_text_entry(self):
        """
            Get currnet text entry value
            @return text as str
        """
        current = None
        if self.__text_entry is not None:
            current = self.__text_entry.value
        return current

    def get_prev_text_entry(self, current_value=None):
        """
            Get previous text entry value
            @param current_value as str
            @return text as str
        """
        previous = None
        if self.__text_entry.prev is not None:
            # Append current to list as it is missing
            if current_value != self.__text_entry.value:
                item = LinkedList(current_value, None, self.__text_entry)
                self.__text_entry.set_next(item)
                previous = self.__text_entry.value
            else:
                current = self.__text_entry
                previous = self.__text_entry.prev.value
                if previous is not None:
                    self.__text_entry = self.__text_entry.prev
                    self.__text_entry.set_next(current)
        return previous

    def get_next_text_entry(self):
        """
            Get next text entry value
            @return text as str
        """
        next = None
        if self.__text_entry.next is not None:
            next = self.__text_entry.next.value
            if next is not None:
                self.__text_entry = self.__text_entry.next
        return next

    def new_page(self, uri, loading_type):
        """
            Open a new page
            @param uri as uri
            @param loading_type as Gdk.LoadingType
        """
        if self.is_ephemeral:
            webview = WebView.new_ephemeral(self.window)
        else:
            webview = WebView.new(self.window)
        webview.show()
        webview.load_uri(uri)
        if loading_type == LoadingType.POPOVER:
            self.window.container.popup_webview(webview)
        else:
            self.window.container.add_webview(webview, loading_type)
            webview.set_parent(self)
            self.add_child(webview)

    def set_parent(self, parent):
        """
            Set webview parent
            @param parent as WebView
        """
        self.__parent = parent

    def set_atime(self, atime):
        """
            Update access time
            @param atime as int
        """
        self.__atime = atime

    def add_child(self, child):
        """
            Add a child to webview
            @param child as WebView
        """
        if child not in self.__children:
            self.__children.insert(0, child)
            child.connect("destroy", self.__on_child_destroy)

    def set_shown(self, shown):
        """
            Set webview as shown
            @param shown as bool
        """
        self.__shown = shown

    def set_window(self, window):
        """
            Update GTK window
            @param window as Window
        """
        self.__window = window

    def stop_loading(self):
        """
            Stop loading webview
        """
        self._loading_state = LoadingState.STOPPED
        WebKit2.WebView.stop_loading(self)

    @property
    def loading_state(self):
        """
            Get loading state
            @return LoadingState
        """
        return self._loading_state

    @property
    def atime(self):
        """
            Get access time
            @return int
        """
        return self.__atime

    @property
    def children(self):
        """
            Get page children
            @return [WebView]
        """
        return self.__children

    @property
    def parent(self):
        """
            Get page parent
            @return WebView/None
        """
        return self.__parent

    @property
    def shown(self):
        """
            True if page already shown on screen (one time)
            @return bool
        """
        return self.__shown

    @property
    def netloc(self):
        """
            Get netloc
            @return str
        """
        uri = self.uri
        parsed = urlparse(uri)
        return parsed.netloc or ""

    @property
    def is_ephemeral(self):
        """
            True if view is private/ephemeral
            @return bool
        """
        return self.get_property("is-ephemeral")

    @property
    def related(self):
        """
            Get related webview
            @return WebView
        """
        return self.__related

    @property
    def window(self):
        """
            Get window
            @return Gtk.Window
        """
        return self.__window

#######################
# PRIVATE             #
#######################
    def __init(self, related, window):
        """
            Init WebView
            @param related as WebView
            @param window as Window
        """
        WebViewHelpers.__init__(self)
        WebViewNightMode.__init__(self)
        WebViewState.__init__(self)
        WebViewErrors.__init__(self)
        WebViewNavigation.__init__(self)
        WebViewSignals.__init__(self)
        WebViewArtwork.__init__(self)
        WebViewCredentials.__init__(self)
        self.__window = window
        self.__atime = 0
        self._loading_state = LoadingState.NONE
        self.__children = []
        self.__parent = None
        self._title = None
        self.__related = related
        self.__shown = False
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.clear_text_entry()
        content_manager = self.get_user_content_manager()
        for content_filter in App().content_filters:
            if content_filter is not None:
                content_manager.add_filter(content_filter)
        if related is None:
            # Set settings
            settings = self.get_settings()
            settings.set_property("enable-java",
                                  App().settings.get_value('enable-plugins'))
            settings.set_property("enable-plugins",
                                  App().settings.get_value('enable-plugins'))
            settings.set_property("minimum-font-size",
                                  App().settings.get_value(
                                      "min-font-size").get_int32())
            if App().settings.get_value("use-system-fonts"):
                self.__set_system_fonts(settings)
            else:
                settings.set_property("monospace-font-family",
                                      App().settings.get_value(
                                          "font-monospace").get_string())
                settings.set_property("sans-serif-font-family",
                                      App().settings.get_value(
                                          "font-sans-serif").get_string())
                settings.set_property("serif-font-family",
                                      App().settings.get_value(
                                          "font-serif").get_string())
            settings.set_property("auto-load-images", True)
            # settings.set_hardware_acceleration_policy(
            #    WebKit2.HardwareAccelerationPolicy.NEVER)
            settings.set_property("enable-site-specific-quirks", True)
            settings.set_property("allow-universal-access-from-file-urls",
                                  False)
            settings.set_property("allow-file-access-from-file-urls", False)
            settings.set_property("enable-javascript", True)
            settings.set_property("enable-media-stream", True)
            settings.set_property("enable-mediasource", True)
            autoplay_videos = App().settings.get_value('autoplay-videos')
            # Setting option to False make video playback buggy
            if not autoplay_videos:
                settings.set_property("media-playback-requires-user-gesture",
                                      True)
            if App().settings.get_value("developer-extras"):
                settings.set_property("enable-developer-extras", True)
                settings.set_enable_write_console_messages_to_stdout(
                    App().settings.get_value("debug"))
            settings.set_property("enable-offline-web-application-cache", True)
            settings.set_property("enable-page-cache", True)
            settings.set_property("enable-resizable-text-areas", True)
            settings.set_property(
                "enable-smooth-scrolling",
                App().settings.get_value("enable-smooth-scrolling"))
            settings.set_property("enable-webaudio", True)
            settings.set_property("enable-webgl", True)
            settings.set_property("javascript-can-access-clipboard", True)
            settings.set_property("javascript-can-open-windows-automatically",
                                  True)
            settings.set_property("media-playback-allows-inline", True)
            if WebKit2.get_minor_version() > 22:
                settings.set_property(
                    "enable-back-forward-navigation-gestures", True)
        self.connect("create", self.__on_create)
        self.connect("load-changed", self._on_load_changed)
        self.connect("notify::is-playing-audio",
                     self.__on_notify_is_playing_audio)
        self.connect("destroy", self.__on_destroy)

    def __set_system_fonts(self, settings):
        """
            Set system font
            @param settings as WebKit2.Settings
            @param system as Gio.Settings("org.gnome.desktop.interface")
        """
        system = Gio.Settings.new("org.gnome.desktop.interface")
        settings.set_property(
            "monospace-font-family",
            system.get_value("monospace-font-name").get_string())
        settings.set_property(
            "sans-serif-font-family",
            system.get_value("document-font-name").get_string())
        settings.set_property(
            "serif-font-family",
            system.get_value("font-name").get_string())

    def __on_create(self, related, navigation_action):
        """
            Create a new view for action
            @param related as WebView
            @param navigation_action as WebKit2.NavigationAction
        """
        webview = WebView.new_with_related_view(related, self.window)
        webview.set_atime(related.atime - 1)
        elapsed = time() - related._last_click_time
        uri = navigation_action.get_request().get_uri()
        webview.connect("ready-to-show",
                        self.__on_ready_to_show,
                        related,
                        navigation_action,
                        elapsed)
        webview.set_uri(uri)
        webview.set_parent(self)
        webview.show()
        self.add_child(webview)
        return webview

    def __on_ready_to_show(self, webview, related, navigation_action, elapsed):
        """
            Add a new webview with related
            @param webview as WebView
            @param related as WebView
            @param navigation_action as WebKit2.NavigationAction
            @param elapsed as float
        """
        properties = webview.get_window_properties()
        if (properties.get_toolbar_visible() or
                properties.get_scrollbars_visible()):
            self.window.container.add_webview(
                webview,
                LoadingType.FOREGROUND)
        else:
            self.window.container.popup_webview(webview)

    def __on_destroy(self, webview):
        """
            Prevent GC
            @param webview as WebView
        """
        self.__parent = None
        self.__related = None
        for child in self.__children:
            child.disconnect_by_func(self.__on_child_destroy)

    def __on_notify_is_playing_audio(self, webview, param):
        """
            Send custom signal
            @param webview as WebView
            @param param as GParam
        """
        value = App().websettings.get("audio", self.uri)
        if value:
            emit_signal(self, "is-playing-audio", self.is_playing_audio())
        else:
            emit_signal(self, "is-playing-audio", False)

    def __on_child_destroy(self, webview):
        """
            Remove from children
            @param webview as WebView
        """
        if webview in self.__children:
            self.__children.remove(webview)


class WebViewMeta(WebViewNavigation, WebView, WebViewErrors,
                  WebViewSignals, WebViewArtwork, WebViewState,
                  WebViewCredentials, WebViewHelpers, WebViewNightMode):

    def __init__(self):
        pass

#######################
# PROTECTED           #
#######################
    def _on_load_changed(self, webview, event):
        """
            Update internals
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        WebViewCredentials._on_load_changed(self, webview, event)
        WebViewHelpers._on_load_changed(self, webview, event)
        WebViewNavigation._on_load_changed(self, webview, event)
        WebViewArtwork._on_load_changed(self, webview, event)
        WebViewNightMode._on_load_changed(self, webview, event)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from locale import strcoll
from time import time

from eolie.define import App
from eolie.widget_bookmark_rating import BookmarkRatingWidget


class TagWidget(Gtk.FlowBoxChild):
    """
        Tag widget with some visual effects
    """

    def __init__(self, title, bookmark_id):
        """
            Init widget
            @param title as str
            @param bookmark_id as int
        """
        Gtk.FlowBoxChild.__init__(self)
        self.__bookmark_id = bookmark_id
        self.__title = title
        grid = Gtk.Grid()
        grid.show()
        grid.get_style_context().add_class("linked")
        self.__entry = Gtk.Entry()
        self.__entry.get_style_context().add_class("tag")
        self.__entry.show()
        self.__entry.set_text(title)
        self.__button = Gtk.Button.new_from_icon_name("window-close-symbolic",
                                                      Gtk.IconSize.BUTTON)
        self.__button.show()
        self.__button.get_style_context().add_class("tag")
        grid.add(self.__entry)
        grid.add(self.__button)
        self.add(grid)
        self.__entry.connect("changed", self.__on_entry_changed)
        self.__button.connect("clicked", self.__on_button_clicked)

    @property
    def label(self):
        """
            Get label
            @return str
        """
        return self.__entry.get_text()

#######################
# PRIVATE             #
#######################
    def __on_button_clicked(self, button):
        """
            Save/Remove tag
            @param button as Gtk.Button
        """
        if button.get_style_context().has_class("suggested-action"):
            title = self.__entry.get_text()
            App().bookmarks.rename_tag(self.__title, title)
            # Update mtime for all tagged bookmarks
            if App().sync_worker is not None:
                mtime = round(time(), 2)
                tag_id = App().bookmarks.get_tag_id(title)
                if tag_id is not None:
                    for (bookmark_id, bookmark_uri, bookmark_title) in\
                            App().bookmarks.get_bookmarks(tag_id):
                        App().bookmarks.set_mtime(bookmark_id, mtime + 1)
                        App().sync_worker.push_bookmark(bookmark_id)
            self.__entry.set_text(title)
            self.__title = title
            self.__on_entry_changed(self.__entry)
        else:
            tag_title = self.__entry.get_text()
            tag_id = App().bookmarks.get_tag_id(tag_title)
            if tag_id is not None:
                App().bookmarks.del_tag_from(tag_id, self.__bookmark_id)
            App().bookmarks.clean_tags()
            self.destroy()

    def __on_entry_changed(self, entry):
        """
            Update button state
            @param entry as Gtk.Entry
        """
        self.__button.set_sensitive(True)
        style_context = self.__button.get_style_context()
        image = self.__button.get_image()
        title = entry.get_text()
        tag_id = App().bookmarks.get_tag_id(title)
        if title == self.__title:
            style_context.remove_class("suggested-action")
            image.set_from_icon_name("window-close-symbolic",
                                     Gtk.IconSize.BUTTON)
        elif tag_id is not None or not title:
            self.__button.set_sensitive(False)
            image.set_from_icon_name("dialog-error-symbolic",
                                     Gtk.IconSize.BUTTON)
        else:
            style_context.add_class("suggested-action")
            image.set_from_icon_name("object-select-symbolic",
                                     Gtk.IconSize.BUTTON)


class BookmarkEditWidget(Gtk.Bin):
    """
        Widget allowing to edit a bookmark
    """

    def __init__(self, bookmark_id, back_enabled=True):
        """
            Init widget
            @param bookmark id as int
            @param enable back button as bool
        """
        Gtk.Bin.__init__(self)
        self.__bookmark_id = bookmark_id
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/BookmarkEdit.ui")
        builder.connect_signals(self)
        self.__flowbox = builder.get_object("flowbox")
        self.__flowbox.set_sort_func(self.__sort_tags)
        self.__add_tag_button = builder.get_object("add_tag_button")
        self.__rename_tag_button = builder.get_object("rename_tag_button")
        self.__remove_tag_button = builder.get_object("remove_tag_button")
        self.__title_entry = builder.get_object("title_entry")
        self.__uri_entry = builder.get_object("uri_entry")
        self.__title_entry.set_text(App().bookmarks.get_title(bookmark_id))
        self.__uri_entry.set_text(App().bookmarks.get_uri(bookmark_id))
        builder.get_object("startup_button").set_active(
            App().bookmarks.get_startup(bookmark_id))
        self.__new_tag_entry = builder.get_object("new_tag_entry")
        # Init new tag completion model
        self.__completion_model = Gtk.ListStore(str)
        self.__completion = Gtk.EntryCompletion.new()
        self.__completion.set_model(self.__completion_model)
        self.__completion.set_text_column(0)
        self.__completion.set_inline_completion(False)
        self.__completion.set_popup_completion(True)
        self.__new_tag_entry.set_completion(self.__completion)
        for (tag_id, title) in App().bookmarks.get_all_tags():
            self.__completion_model.append([title])

        for title in App().bookmarks.get_tags(bookmark_id):
            tag = TagWidget(title, bookmark_id)
            tag.show()
            self.__flowbox.add(tag)
        if not back_enabled:
            builder.get_object("back_button").hide()
        bookmark_rating = BookmarkRatingWidget(bookmark_id)
        bookmark_rating.show()
        builder.get_object("bookmark_grid").attach(bookmark_rating, 2, 1, 1, 1)
        self.add(builder.get_object("widget"))
        self.connect("unmap", self.__on_unmap)
        self.__updated = False

#######################
# PROTECTED           #
#######################
    def _on_back_clicked(self, button):
        """
            Destroy self
            @param button as Gtk.Button
        """
        GLib.idle_add(self.hide)
        GLib.timeout_add(2000, self.destroy)

    def _on_del_clicked(self, button):
        """
            Remove item
            @param button as Gtk.Button
        """
        self.disconnect_by_func(self.__on_unmap)
        if App().sync_worker is not None:
            guid = App().bookmarks.get_guid(self.__bookmark_id)
            App().sync_worker.remove_from_bookmarks(guid)
        App().bookmarks.remove(self.__bookmark_id)
        if isinstance(self.get_parent(), Gtk.Popover):
            self.get_parent().hide()
        else:
            self.get_parent().set_visible_child_name("bookmarks")

    def _on_new_tag_entry_activate(self, entry, ignore1=None, ignore2=None):
        """
            Add new tag
            @param entry as Gtk.Entry
        """
        tag_title = self.__new_tag_entry.get_text()
        if not tag_title:
            return
        if not App().bookmarks.has_tag(self.__bookmark_id, tag_title):
            tag_id = App().bookmarks.get_tag_id(tag_title)
            if tag_id is None:
                tag_id = App().bookmarks.add_tag(tag_title)
            App().bookmarks.add_tag_to(tag_id, self.__bookmark_id)
            tag = TagWidget(tag_title, self.__bookmark_id)
            tag.show()
            self.__flowbox.add(tag)
        entry.set_text("")

    def _on_flowbox_size_allocate(self, scrolled, allocation):
        """
            Set scrolled size allocation based on viewport allocation
            @param scrolled as Gtk.ScrolledWindow
            @param flowbox allocation as Gtk.Allocation
        """
        height = allocation.height
        if height > 300:
            height = 300
        scrolled.set_size_request(-1, height)

    def _on_entry_changed(self, entry):
        """
            Mark bookmark as updated
            @param entry as Gtk.Entry
        """
        self.__updated = True

    def _on_load_at_startup_toggled(self, button):
        """
            Set bookmark startup state
            @param button as Gtk.ToggleButton
        """
        App().bookmarks.set_startup(self.__bookmark_id, button.get_active())

#######################
# PRIVATE             #
#######################
    def __sort_tags(self, child1, child2):
        """
            Sort tags
            @param child1 as TagWidget
            @param child2 as TagWidget
        """
        return strcoll(child1.label, child2.label)

    def __on_unmap(self, widget):
        """
            Save uri and title
            @param widget as Gtk.Widget
        """
        if self.__updated:
            App().bookmarks.set_title(self.__bookmark_id,
                                      self.__title_entry.get_text())
            App().bookmarks.set_uri(self.__bookmark_id,
                                    self.__uri_entry.get_text())
            App().bookmarks.clean_tags()
            if App().sync_worker is not None:
                App().bookmarks.set_mtime(self.__bookmark_id,
                                          round(time(), 2) + 1)
                App().sync_worker.push_bookmark(self.__bookmark_id)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from eolie.define import App


class WebViewStateStruct:
    def __init__(self):
        self.uri = ""
        self.title = ""
        self.atime = 0
        self.is_ephemeral = False
        self.session = None


class WebViewState:
    """
        Web view state allowing to restore a webview from disk
    """

    def new_from_state(state, window):
        """
            New webview from state
            @param state as WebViewStateStruct
            @param window as Window
        """
        from eolie.webview import WebView
        if state.is_ephemeral:
            webview = WebView.new_ephemeral(window)
        else:
            webview = WebView.new(window)
        webview.set_uri(state.uri)
        webview.set_title(state.title)
        webview.set_atime(state.atime)
        # TODO session
        return webview

    def __init__(self):
        """
            Init state
        """
        pass

    @property
    def state(self):
        """
            Get state
            @return WebViewStateStruct
        """
        if App().settings.get_value("remember-session"):
            state = WebViewStateStruct()
            state.uri = self.uri
            state.title = self.title
            state.atime = self.atime
            state.is_ephemeral = self.is_ephemeral
            state.session = self.get_session_state().serialize().get_data()
            return state
        else:
            return None



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, WebKit2, Pango

from gettext import gettext as _

from eolie.define import App
from eolie.popover_downloads import DownloadsPopover


class ProgressBar(Gtk.ProgressBar):
    """
        Simple progress bar with width contraint
    """

    def __init__(self):
        Gtk.ProgressBar.__init__(self)
        self.set_property("valign", Gtk.Align.END)
        self.get_style_context().add_class("progressbar-button")

    def do_get_preferred_width(self):
        return (24, 24)


class ToolbarEnd(Gtk.Bin):
    """
        Toolbar end
    """

    def __init__(self, window, fullscreen):
        """
            Init toolbar
            @param window as Window
            @param fullscreen as bool
        """
        Gtk.Bin.__init__(self)
        self.__window = window
        self.__timeout_id = None
        self.__image_change_state_id = None
        self.set_hexpand(True)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/ToolbarEnd.ui")
        builder.connect_signals(self)
        self.__home_button = builder.get_object("home_button")
        self.__menu_button = builder.get_object("menu_button")
        self.__download_button = builder.get_object("download_button")
        self.__settings_button = builder.get_object("settings_button")
        self.__tls_button = builder.get_object("tls_button")
        if fullscreen:
            builder.get_object("fullscreen_button").show()
        self.__progress = ProgressBar()
        if App().download_manager.get():
            self.__progress.show()
        App().download_manager.connect("download-start",
                                       self.__on_download)
        App().download_manager.connect("download-finish",
                                       self.__on_download)
        overlay = builder.get_object("overlay")
        overlay.add_overlay(self.__progress)
        overlay.set_overlay_pass_through(self.__progress, True)
        self.add(builder.get_object("end"))

    def show_tls_button(self, show):
        """
            Show TLS button allowing user to discard a certificate
            @param show as bool
        """
        if show:
            self.__tls_button.show()
        else:
            self.__tls_button.hide()

    def show_download(self, download):
        """
            Notify user about download
            @param download as WebKit2.Download
        """
        header = Gtk.Label.new()
        header.set_markup("<b>" + _("Downloading:") + "</b>")
        header.set_ellipsize(Pango.EllipsizeMode.END)
        header.show()
        destination = download.get_destination()
        try:
            uri = Gtk.Label.new(destination.split("/")[-1])
        except:
            uri = Gtk.Label.new(download.get_destination())
        uri.set_max_width_chars(30)
        uri.set_ellipsize(Pango.EllipsizeMode.END)
        uri.show()
        grid = Gtk.Grid()
        grid.set_margin_start(5)
        grid.set_margin_end(5)
        grid.set_margin_top(5)
        grid.set_margin_bottom(5)
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.show()
        grid.add(header)
        grid.add(uri)
        popover = Gtk.Popover.new()
        popover.get_style_context().add_class("dark")
        popover.add(grid)
        popover.set_modal(False)
        popover.set_relative_to(self.__download_button)
        popover.popup()
        GLib.timeout_add(2000, popover.destroy)

    def save_page(self):
        """
            Show a dialog allowing user to save current page
        """
        filechooser = Gtk.FileChooserNative.new(_("Save page"),
                                                self.__window,
                                                Gtk.FileChooserAction.SAVE,
                                                _("Save"),
                                                _("Cancel"))
        filechooser.connect("response", self.__on_save_response)
        filechooser.run()

    @property
    def home_button(self):
        """
            Get home button
            @return Gtk.ToogleButton
        """
        return self.__home_button

    @property
    def menu_button(self):
        """
            Get download button
            @return Gtk.ToogleButton
        """
        return self.__menu_button

    @property
    def download_button(self):
        """
            Get download button
            @return Gtk.ToogleButton
        """
        return self.__download_button

#######################
# PROTECTED           #
#######################
    def _on_download_button_toggled(self, button):
        """
            Show download popover
            @param button as Gtk.Button
        """
        self.__window.close_popovers()
        if button == self.__download_button and not button.get_active():
            return
        popover = DownloadsPopover(self.__window)
        # We are relative to toolbar button, button can be in menu
        popover.set_relative_to(button)
        popover.connect("closed", self.__on_popover_closed, button)
        popover.set_modal(False)
        self.__window.register(popover)
        popover.popup()

    def _on_tls_button_clicked(self, button):
        """
            Do not accept TLS for this site by default
            @param button as Gtk.Button
        """
        button.hide()
        uri = self.__window.container.webview.uri
        App().websettings.set("accept_tls", uri, False)
        self.__window.container.close_webview(self.__window.container.webview)

    def _on_fullscreen_button_clicked(self, button):
        """
            Leave fullscreen mode
            @param button as Gtk.Button
        """
        self.__window.unfullscreen()

    def _on_home_button_clicked(self, button):
        """
            Go to home page
            @param button as Gtk.Button
        """
        self.__window.close_popovers()
        self.__window.container.webview.load_uri(App().start_page)

    def _on_menu_button_toggled(self, button):
        """
            Show settings menu
            @param button as Gtk.ToogleButton
        """
        self.__window.close_popovers()
        uri = self.__window.container.webview.uri
        if not button.get_active() or not uri:
            return
        from eolie.menu_toolbar import ToolbarMenu
        popover = ToolbarMenu(uri, self.__window)
        popover.set_relative_to(button)
        popover.set_modal(False)
        self.__window.register(popover)
        popover.connect("closed", self.__on_popover_closed, button)
        popover.popup()

#######################
# PRIVATE             #
#######################
    def __hide_progress(self):
        """
            Hide progress if needed
        """
        if self.__timeout_id is None:
            self.__progress.hide()

    def __update_progress(self, download_manager):
        """
            Update progress
        """
        fraction = 0.0
        nb_downloads = 0
        for download in download_manager.get():
            nb_downloads += 1
            fraction += download.get_estimated_progress()
        if nb_downloads:
            value = fraction / nb_downloads
            self.__progress.set_fraction(value)
            App().update_unity_badge(value)
        return True

    def __on_save_response(self, dialog, response_id):
        """
            Tell WebKit to save current page
            @param dialog as Gtk.NativeDialog
            @param response_id as int
        """
        if response_id == Gtk.ResponseType.ACCEPT:
            self.__window.container.webview.save_to_file(
                dialog.get_file(),
                WebKit2.SaveMode.MHTML,
                None,
                None)

    def __on_download(self, download_manager, name=""):
        """
            Update progress bar
            @param download_manager as DownloadManager
            @param name as str (do not use this)
        """
        if download_manager.active:
            # Show button if needed
            if not self.__download_button.is_visible() and\
                    self.__home_button.is_visible():
                self.__download_button.show()
            if self.__timeout_id is None:
                self.__progress.show()
                self.__timeout_id = GLib.timeout_add(1000,
                                                     self.__update_progress,
                                                     download_manager)
        elif self.__timeout_id is not None:
            self.__progress.set_fraction(1.0)
            App().update_unity_badge(1.0)
            GLib.timeout_add(1000, self.__hide_progress)
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None
            self.__download_button.get_style_context().add_class("selected")

    def __on_popover_closed(self, popover, button):
        """
            Unlock focus
            @param popover as Gtk.Popover
            @param button as Gtk.Button
        """
        button.get_style_context().remove_class("selected")
        button.set_active(False)
        if self.__image_change_state_id is not None:
            self.__images_action.disconnect(self.__image_change_state_id)
            self.__image_change_state_id = None
